<?php
/**
 * @file
 * Functions used on ftp-connected environments during deploys.
 */

/**
 * Get files on remote server.
 *
 * @param resource $server
 *   FTP resource.
 * @param array $gets
 *   Array containing ('CRIT'|'WARN', [remote file], [local file]).
 *   If array[0] == 'CRIT', we will set $deploy_context['failed'] to 1.
 * @param array $deploy_context
 *   Deploy context.
 */
function _deployer_ftp_get(&$server, $gets, &$deploy_context = NULL) {
  foreach ($gets as $get) {
    $remote = $get[1];
    $local = $get[2];

    $size = _deployer_ftp_cmd('ftp_size', $server, array($remote));
    if ($size !== FALSE) {
      $size = format_size($size);
    }
    else {
      $size = t('Uknown size');
    }

    _deployer_log(t('Downloading %f% (%size%) from %s% to %l%: ', array(
      '%f%' => $remote,
      '%s%' => $server->hostname,
      '%size%' => $size,
      '%l%' => $local,
    )));

    $mode = FTP_BINARY;

    $transportation = _deployer_ftp_cmd('ftp_get', $server, array(
      $local,
      $remote,
      $mode,
    ));

    if (!$transportation) {
      _deployer_log('transport failed');
    }
    if (!filesize($local)) {
      _deployer_log('filesize is NOT');
    }

    if (!$transportation || !filesize($local)) {
      _deployer_log(t('   ==> failed. (CRITICAL)'));
      if ($get[0] == 'CRIT' && !empty($deploy_context)) {
        $deploy_context['failed'] = 1;
      }
    }
    else {
      _deployer_log(t('   ==> ok'));
    }
  }
}

/**
 * Download files / folders recursively.
 *
 * @param object $server
 *   Server object with ftp resource.
 *
 * @param string $remote
 *   Remote path/file to download.
 *
 * @param string $local_path
 *   Local path to download into.
 *
 * @param int $echo
 *   Echo output (0 = no, 1 = yes)
 *
 * @param int $echo_recursive
 *   Echo output for recursive transfers.
 *
 * @param array $deploy_context
 *   Deploy context.
 *
 *  * @param int $count
 *   Count files transferred.
 *
 * @param int $total
 *   Total amount of files to transfer.
 *
 * @param string $label
 *   Label to use for progress placeholder.
 *
 * @param int $iteration
 *   Tracks first - or other - iteration.
 */
function _deployer_ftp_get_recursive(&$server, $remote, $local_path, $echo, $echo_recursive, &$deploy_context = NULL, &$count = -1, $total = 0, $label = '---', $iteration = 0) {

  if (!$deploy_context['failed']) {

    $old_pwd = _deployer_ftp_cmd('ftp_pwd', $server, array());

    if (!$iteration) {
      $label = uniqid();
      $label_dir_size = uniqid();
      _deployer_set_placeholder($label_dir_size, t(' - calculating size - '));
      _deployer_set_placeholder($label, 0);
    }

    $iteration++;

    if ($count > 0 && $total > 0) {
      $percentage = ceil($count / $total * 100);
      _deployer_set_placeholder($label, $percentage);
    }

    if ($remote != '.' && $remote != '..') {
      $count++;

      $size = _deployer_ftp_cmd('ftp_size', $server, array($remote));
      if ($size === FALSE) {
        // File doesnt exist.
        if ($echo) {
          _deployer_log(t('File does not exist (might have disappeared: @f', array('@f' => $remote)));
        }
      }
      elseif ($size == '-1') {
        // Is directory.
        drupal_mkdir($local_path . '/' . $remote, NULL, TRUE);

        if ($echo) {
          if ($iteration === 1) {
            // Feedback.
            _deployer_log(t('  @r (@s)', array('@r' => $remote . '/', '@s' => $label_dir_size)));

            // Get dir size from _deployer_ftp_dirlist_recursive();
            $dirlist = _deployer_ftp_dirlist_recursive($server, $remote);
            $num_files = count($dirlist);
            unset($dirlist);
            $total = $num_files;

            global $_deployer_dirsize;
            $dirsize = format_size($_deployer_dirsize);
            _deployer_set_placeholder($label_dir_size, $dirsize . ' in ' . $total . ' files');

          }

          _deployer_log(t('       Progress: !p %', array('!p' => $label)));
        }

        // Recurse into directory.
        $status = _deployer_ftp_cmd('ftp_chdir', $server, array($remote));
        if (!$status) {
          if ($echo) {
            _deployer_log(t('    => failed!'));
          }
          if (!empty($deploy_context)) {
            $deploy_context['failed'] = 1;
          }
        }
        else {
          $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
          if (!empty($contents)) {
            foreach ($contents as $file) {
              // $file can also be a folder.
              if (!$deploy_context['failed']) {
                _deployer_ftp_get_recursive($server, $file, $local_path . '/' . $remote, $echo_recursive, $echo_recursive, $deploy_context, $count, $total, $label, $iteration);
              }
            }
          }
        }

        // Go back up.
        _deployer_ftp_cmd('ftp_chdir', $server, array($old_pwd));
      }
      else {
        // Is file.
        $size = format_size($size);
        $mode = FTP_BINARY;
        drupal_mkdir($local_path, NULL, TRUE);
        if ($echo) {
          _deployer_log(t('  %f% (%size%) => %l%: ', array(
            '%f%' => $remote,
            '%size%' => $size,
            '%l%' => $local_path . '/',
          )));
        }

        $transportation = _deployer_ftp_cmd('ftp_get', $server, array(
          $local_path . '/' . basename($remote),
          $remote,
          $mode,
        ));

        if ($transportation) {
          if ($echo) {
            _deployer_log(t('    => ok'));
          }
        }
        else {
          if ($echo) {
            _deployer_log(t('    => failed!'));
          }
          if (!empty($deploy_context)) {
            // Check connectivity.
            // If webroot folder is available and if file size is FALSE,
            // assume file has been removed in the mean time.
            if (_deployer_ftp_cmd('ftp_size', $server, array($server->webroot)) !== FALSE && _deployer_ftp_cmd('ftp_size', $server, array($remote)) === FALSE) {
              _deployer_log(t('(WARNING) File was removed in the meantime'));
            }
            else {
              $deploy_context['failed'] = 1;
            }
          }
        }
      }
    }
  }

  if ($count > 0 && $total > 0) {
    $percentage = ceil($count / $total * 100);
    _deployer_set_placeholder($label, $percentage);
  }

}

/**
 * List files in a folder recursively.
 *
 * After each call, global param int $_deployer_dirsize
 *   gives the total size in bytes of the folder structure.
 *
 * @param object $server
 *   Server object with ftp resource.
 *
 * @param string $path
 *   Remote path to list content of.
 *
 * @param int $depth
 *   Depth to list. Default = 0 (infinite).
 *   1 means list content in current dir, no recursion at all.
 *
 * @param string $sub_path
 *   Keeps track of where we are 'now'. No manual use.
 *
 * @param int $cur_depth
 *   Keeps track of current depth.
 *
 * @param int $iteration
 *   Tracks if this is the first or other iteration.
 *
 * @return array
 *   File list array.
 */
function _deployer_ftp_dirlist_recursive(&$server, $path, $depth = 0, $sub_path = '', $cur_depth = 0, $iteration = 0) {
  global $_deployer_dirsize;

  // Define dirsize var.
  if (!$iteration && isset($_deployer_dirsize)) {
    $_deployer_dirsize = 0;
  }

  $iteration++;

  $file_array = array();
  $old_pwd = _deployer_ftp_cmd('ftp_pwd', $server, array());

  if ($path != '.' && $path != '..') {
    $size = _deployer_ftp_cmd('ftp_size', $server, array($path));

    // Track directory size.
    $_deployer_dirsize = $_deployer_dirsize + $size;

    if ($size === FALSE) {
      // This is no file nor a folder. WTF?
      _deployer_log('FALSE ??');
    }
    elseif ($size == '-1') {
      // Is directory.
      // Recurse into directory.
      $status = _deployer_ftp_cmd('ftp_chdir', $server, array($path));

      if ($status) {
        // Set current path and append to file array.
        if (empty($sub_path)) {
          $sub_path = '/';
        }
        else {
          $sub_path = $sub_path . '/' . $path;
          // Remove leading '/'.
          $sub_path = preg_replace('|^/+|', '', $sub_path);
          $file_array[] = $sub_path;
          $cur_depth++;
        }

        // Check for depth and recurse.
        if ($cur_depth < $depth || $depth == 0) {
          $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
          if (!empty($contents)) {
            foreach ($contents as $file) {
              // $file can also be a folder.
              $file_array = array_merge($file_array, _deployer_ftp_dirlist_recursive($server, $file, $depth, $sub_path, $cur_depth, $iteration));
            }
          }
        }
      }

      // Go back up.
      _deployer_ftp_cmd('ftp_chdir', $server, array($old_pwd));
    }
    else {
      // Is file.
      // Remove leading '/'.
      $file_path = $sub_path . '/' . $path;
      $file_path = preg_replace('|^/+|', '', $file_path);
      $file_array[] = $file_path;
    }
  }

  return $file_array;
}

/**
 * Put file on server.
 *
 * @param object $server
 *   Server object with ftp resource.
 *
 * @param string $remote
 *   Remote filename.
 *
 * @param string $local
 *   Local path/file to put on server.
 *
 * @param int $echo
 *   Echo output (0 = no, 1 = yes)
 *
 * @param array $deploy_context
 *   Deploy context.
 */
function _deployer_ftp_put_file(&$server, $remote, $local, $echo, &$deploy_context = NULL) {
  if (!$deploy_context['failed']) {

    _deployer_set_placeholder('&&STATUS_' . $local . '&&', 'uploading');
    $size = format_size(filesize($local));
    $mode = FTP_BINARY;

    // Create containing folder if not existing.
    $old_pwd = _deployer_ftp_cmd('ftp_pwd', $server, array());

    $dir = dirname($remote);
    if ($dir && $dir != NULL && !empty($dir) && $dir != $old_pwd) {
      if (!_deployer_ftp_cmd('ftp_chdir', $server, array(dirname($remote)), FALSE)) {
        _deployer_ftp_cmd('ftp_mkdir', $server, array(dirname($remote)));
        _deployer_ftp_cmd('ftp_chmod', $server, array(0755, dirname($remote)));
      }
      else {
        _deployer_ftp_cmd('ftp_chdir', $server, array($old_pwd));
      }
    }

    if ($echo) {
      _deployer_log(t('  !s: %f% (%size%) => %l%: ', array(
        '!s' => '&&STATUS_' . $local . '&&',
        '%f%' => $local,
        '%size%' => $size,
        '%l%' => $remote,
      )));
    }
    $transportation = _deployer_ftp_cmd('ftp_put', $server, array(
      $remote,
      $local,
      $mode,
    ));
    if ($transportation) {
      _deployer_ftp_cmd('ftp_chmod', $server, array(0644, $remote));
      if ($echo) {
        _deployer_set_placeholder('&&STATUS_' . $local . '&&', 'done');
      }
    }
    else {
      if ($echo) {
        _deployer_set_placeholder('&&STATUS_' . $local . '&&', 'FAILED');
      }
      if (!empty($deploy_context)) {
        $deploy_context['failed'] = 1;
      }
    }
  }
}

/**
 * Put files / folders recursively.
 *
 * @param object $server
 *   Server object with ftp resource.
 *
 * @param string $remote_path
 *   Remote path to put files into.
 *
 * @param string $local
 *   Local path/file to put on server.
 *
 * @param int $echo
 *   Echo output (0 = no, 1 = yes)
 *
 * @param int $echo_recursive
 *   Echo output for recursive transfers.
 *
 * @param array $deploy_context
 *   Deploy context.
 *
 * @param array $excludes
 *   Full-path files/folders to exclude from transfer.
 *
 * @param int $count
 *   Count files transferred.
 *
 * @param int $total
 *   Total amount of files to transfer.
 *
 * @param string $label
 *   Label to use for progress placeholder.
 *
 * @param int $dirsize
 *   Track size of current dir.
 */
function _deployer_ftp_put_recursive(&$server, $remote_path, $local, $echo, $echo_recursive, &$deploy_context = NULL, $excludes = array(), &$count = 0, $total = 0, $label = '---', $dirsize = 0) {

  if (!$deploy_context['failed']) {
    $old_pwd = _deployer_ftp_cmd('ftp_pwd', $server, array());

    if ($count < 1) {
      // Get amount of files to transport.
      exec('/usr/bin/find ' . $local . ' | wc -l', $output, $return_stat);
      if (is_numeric($output[0])) {
        $total = $output[0];
        $label = $local;
        _deployer_set_placeholder('&&PROGRESS_' . $label . '&&', 0);

        // Get directory size to transport.
        unset($output);
        unset($return_stat);
        exec('du -sb ' . $local, $output, $return_stat);
        if ($return_stat == 0) {
          $dirsize = $output[0];
        }
      }
      else {
        _deployer_log(t('Cannot determine number of files: @o', array('@o' => $output)));
      }
    }

    if ($count > 0 && $total > 0) {
      $percentage = ceil($count / $total * 100);
      _deployer_set_placeholder('&&PROGRESS_' . $label . '&&', $percentage);
    }

    if (!in_array($local, $excludes)) {
      if (basename($local) != '.' && basename($local) != '..') {
        if (!is_dir($local) && !is_file($local)) {
          // File doesnt exist.
          _deployer_log(t('Local file or directory @f does not exist.', array('@f' => $local)));
          if (!empty($deploy_context)) {
            $deploy_context['failed'] = 1;
          }
          _deployer_log(t('File does not exist: @f', array('@f' => $remote_path)));
        }
        elseif (is_dir($local)) {
          // Is directory..
          if ($echo) {
            if ($dirsize) {
              $size = '(' . format_size($dirsize) . ')';
            }
            else {
              $size = '';
            }
            _deployer_log(t('  @r @s', array('@r' => $local . '/', '@s' => $size)));
            _deployer_log(t('       Progress: !p %', array('!p' => '&&PROGRESS_' . $label . '&&')));
          }

          // Recurse into directory.
          _deployer_ftp_mksubdirs($server, '/', $remote_path . '/' . basename($local));
          $count++;

          _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $remote_path . '/' . basename($local)));
          $status = _deployer_ftp_cmd('ftp_chdir', $server, array($remote_path . '/' . basename($local)));
          if (!$status) {
            _deployer_log(t('    => failed changing dir to %d!', array('%d' => $remote_path . '/' . basename($local))));
            if (!empty($deploy_context)) {
              $deploy_context['failed'] = 1;
            }
          }
          else {
            $contents = scandir($local);
            if (!empty($contents)) {
              foreach ($contents as $file) {
                // $file can also be a folder.
                if (!$deploy_context['failed']) {
                  _deployer_ftp_put_recursive($server, $remote_path . '/' . basename($local), $local . '/' . $file, $echo_recursive, $echo_recursive, $deploy_context, $excludes, $count, $total, $label);
                  if ($deploy_context['failed']) {
                    _deployer_log(t(' => failed'));
                  }
                }
              }
            }
          }

          // Go back up.
          _deployer_ftp_cmd('ftp_chdir', $server, array($old_pwd));
        }
        else {
          // Is file.
          $size = format_size(filesize($local));
          $mode = FTP_BINARY;

          _deployer_ftp_mksubdirs($server, '/', $remote_path);
          if ($echo) {
            _deployer_log(t('  %f% (%size%) => %l%: ', array(
              '%f%' => $local,
              '%size%' => $size,
              '%l%' => $remote_path . '/',
            )));
          }

          $transportation = _deployer_ftp_cmd('ftp_put', $server, array(
            $remote_path . '/' . basename($local),
            $local,
            $mode,
          ));

          $count++;
          if ($transportation) {
            _deployer_ftp_cmd('ftp_chmod', $server, array(0644, $remote_path . '/' . basename($local)));
            if ($echo) {
              _deployer_log(t('    => ok'));
            }
          }
          else {
            _deployer_log(t('    => transport of @f failed!', array('@f' => $local)));
            if (!empty($deploy_context)) {
              $deploy_context['failed'] = 1;
            }
          }
        }
      }
    }
    else {
      _deployer_log(t(' Excluding %f', array('%f' => $local)));
    }

  }

  if ($count > 0 && $total > 0) {
    $percentage = ceil($count / $total * 100);
    _deployer_set_placeholder('&&PROGRESS_' . $label . '&&', $percentage);
  }
}

/**
 * Create dir recursively.
 *
 * @param object $server
 *   Server object with ftp resource.
 *
 * @param string $remote
 *   Remote path to create $path in. (file or folder).
 *
 * @param string $path
 *   Relative path to create.
 */
function _deployer_ftp_mksubdirs(&$server, $remote, $path) {
  $old_pwd = _deployer_ftp_cmd('ftp_pwd', $server, array());

  _deployer_ftp_cmd('ftp_chdir', $server, array($remote));
  $parts = explode('/', $path);
  foreach ($parts as $part) {
    if (!empty($part)) {
      if (!_deployer_ftp_cmd('ftp_chdir', $server, array($part), FALSE)) {
        _deployer_ftp_cmd('ftp_mkdir', $server, array($part));
        _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $part));
        _deployer_ftp_cmd('ftp_chdir', $server, array($part));
      }
    }
  }

  _deployer_ftp_cmd('ftp_chdir', $server, array($old_pwd));
}

/**
 * Delete files / folders recursively.
 *
 * @param object $server
 *   Server object with ftp resource.
 *
 * @param string $remote
 *   Remote path to delete (file or folder).
 *
 * @param int $echo
 *   Echo output (0 = no, 1 = yes)
 *
 * @param int $echo_recursive
 *   Echo output for recursive transfers.
 *
 * @param array $deploy_context
 *   Deploy context.
 *
 * @param bool $force
 *   Force action, even if context failed.
 */
function _deployer_ftp_delete_recursive(&$server, $remote, $echo, $echo_recursive, &$deploy_context = NULL, $force = FALSE) {
  if (!$deploy_context['failed'] || $force) {
    $old_pwd = _deployer_ftp_cmd('ftp_pwd', $server, array());

    if ($remote != '.' && $remote != '..') {
      $size = _deployer_ftp_cmd('ftp_size', $server, array($remote), FALSE);
      if ($size === FALSE) {
        // File doesnt exist.
        if ($echo) {
          _deployer_log(t('File does not exist: @f (WARNING)', array('@f' => $remote)));
        }
      }
      elseif (_deployer_ftp_cmd('ftp_chdir', $server, array($remote), FALSE)) {
        // Is directory.
        // Recurse into directory.
        $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
        if (!empty($contents)) {
          foreach ($contents as $file) {
            // $file can also be a folder.
            if (!$deploy_context['failed'] || $force) {
              _deployer_ftp_delete_recursive($server, $file, $echo_recursive, $echo_recursive, $deploy_context, TRUE);
            }
          }
        }

        // Go back up.
        _deployer_ftp_cmd('ftp_chdir', $server, array($old_pwd));

        // Remove empty folder.
        if ($echo) {
          _deployer_log(t('  rm @r', array('@r' => $remote . '/')));
        }
        _deployer_ftp_cmd('ftp_rmdir', $server, array($remote));
      }
      elseif ($size != -1) {
        // Is file.
        $size = format_size($size);
        if ($echo) {
          _deployer_log(t('  %f% (%size%)', array('%f%' => $remote, '%size%' => $size)));
        }
        $del = _deployer_ftp_cmd('ftp_delete', $server, array($remote));
        if ($del) {
          if ($echo) {
            _deployer_log(t('    => ok'));
          }
        }
        else {
          if ($echo) {
            _deployer_log(t('    => failed!'));
          }
        }
      }
    }
  }
}

/**
 * RPC action on remote url.
 *
 * @param string $url
 *   Base url, for example 'https://webserver1.com'
 *
 * @param string $query
 *   Query string in format 'param=value&param=value&...'.
 *
 * @param array $deploy_context
 *   Deploy context.
 *
 * @param string $impact
 *   'crit' -> $deploy_context['failed']=1 on failure.
 *   'warn' -> No failure set.
 *
 * @return bool|object
 *   FALSE or drupal_http_request object.
 */
function _deployer_rpc($url, $query, &$deploy_context, $impact = 'crit') {
  $request = $url . '/deploy_rpc.php';
  $query_orig = $query;
  $query = '?key=' . $deploy_context['rpc_key'] . '&' . $query;

  // Set timeout to 4 minutes.
  $options = array(
    'timeout' => 240,
  );

  $response = drupal_http_request($request . $query, $options);

  if (!$response) {
    _deployer_log(t('Failed to connect RPC on @s', array('@s' => $url)));
    if ($impact == 'crit') {
      $deploy_context['failed'] = 1;
    }
    return FALSE;
  }
  elseif (strpos($response->data, 'FALSE') !== FALSE || empty($response->data)) {
    _deployer_log(t('Failed to execute query @q on @s', array('@s' => $url, '@q' => $query_orig)));
    _deployer_log(t('Error') . ': '  . $response->data);
    if ($impact == 'crit') {
      $deploy_context['failed'] = 1;
    }
    return FALSE;
  }

  return $response;
}

/**
 * Post file to an url.
 *
 * @param string $file
 *   File path on filesystem.
 *
 * @param string $url
 *   URL to post to, including query.
 *
 * @param string $ok_string
 *   String to filter for in return data.
 *   If this string is not matched, we assume the call has failed.
 *   Default: 'ok'.
 *
 * @return bool
 *   TRUE|FALSE
 */
function _deployer_ftp_post_file($file, $url, $ok_string = 'ok') {
  if (is_file($file)) {
    $post = array('file' => '@' . $file);

    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $url);
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $post);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
    $result = curl_exec($ch);
    curl_close($ch);

    if (strpos($result, $ok_string) !== FALSE) {
      return TRUE;
    }
    else {
      return FALSE;
    }
  }
  else {
    _deployer_log(t('File %f does not exist!', array('%f' => $file)));
    return FALSE;
  }
}

/**
 * Call ftp command.
 *
 * Try it for couple of times, reconnect on last retry.
 *
 * @param string $command
 *   FTP command. See http://www.php.net/manual/en/ref.ftp.php .
 *
 * @param object $server
 *   Server object, containing ->ftp (resource).
 *   $server->ftp is ALWAYS passed as first argument to $command!
 *
 * @param array $args
 *   Arguments for ftp command.
 *
 * @param bool $failsafe
 *   TRUE: retry for couple of times and reconnect.
 *   FALSE: perform command just once.
 *
 * @return bool|mixed
 *   Return value from command.
 */
function _deployer_ftp_cmd($command, &$server, $args, $failsafe = TRUE) {

  $retval = FALSE;
  $debug_output_done = 0;

  // Set FTP failback properties.
  global $_deployer_ftp_retries , $_deployer_ftp_reconnect_after;
  if (empty($_deployer_ftp_retries) || empty($_deployer_ftp_reconnect_after)) {
    $_deployer_ftp_retries = variable_get('deployer_ftp_max_retries', DEPLOYER_FTP_RETRIES);
    $_deployer_ftp_reconnect_after = variable_get('deployer_ftp_reconnect_after', DEPLOYER_FTP_RECONNECT);
  }

  // Assume the first argument for all ftp commands is $server->ftp.
  $arguments = array(&$server->ftp);
  if (!empty($args)) {
    foreach ($args as $arg) {
      $arguments[] = $arg;
    }
  }
  $args = $arguments;

  $label = '';
  $retry_count = 0;
  while ($retry_count <= $_deployer_ftp_retries) {

    // Give feedback.
    if ($retry_count == 1) {
      // Generate placeholder.
      if (empty($label)) {
        $label = uniqid();
      }
      _deployer_set_placeholder($label, $retry_count);
      _deployer_log(t('Retrying command @c (!r)', array('@c' => $command, '!r' => $label)));
    }
    elseif ($retry_count > 1) {
      _deployer_set_placeholder($label, $retry_count);
    }

    if ($retry_count >= $_deployer_ftp_reconnect_after) {
      // Provide debug output in case of reconnect.
      // Only once.
      if (!$debug_output_done) {
        _deployer_log(t('Debug output for this failing command'));
        $debug_output_done = 1;
        $backtrace = debug_backtrace(NULL, 4);
        foreach ($backtrace as $item) {
          _deployer_log("\n");
          if (isset($item['file'])) {
            // Make path relative.
            $file = str_replace($_SERVER['DOCUMENT_ROOT'] . $GLOBALS['base_path'], '', $item['file']);
            _deployer_log('  ' . t('> file: @f', array('@f' => $file)));
          }
          if (isset($item['line'])) {
            _deployer_log('  ' . t('> line: @l', array('@l' => $item['line'])));
          }
          if (isset($item['function'])) {
            _deployer_log('  ' . t('> function: @f', array('@f' => $item['function'])));
          }
        }
      }

      // Last resort, reconnect.
      $newserver = deployer_reconnect_ftp($server);
      if ($newserver) {
        $server = $newserver;
        $args[0] = &$server->ftp;
      }
    }

    $retval = call_user_func_array($command, $args);
    if ($retval !== FALSE) {
      // Track current folder.
      if ($command == 'ftp_chdir') {
        $server->current_dir = _deployer_ftp_cmd('ftp_pwd', $server, array());
      }

      // Set feedback.
      if ($retry_count > 0) {
        _deployer_set_placeholder($label, t('@r => ok', array('@r' => $retry_count)));
      }
      break;
    }
    elseif (!$failsafe) {
      break;
    }

    $retry_count++;
  }

  return $retval;
}

/**
 * Build server objects.
 *
 * @param object $deploynode
 *   Node object.
 * @param array $deploy_context
 *   Context array.
 * @param array $servers
 *   Servers array containing server objects.
 */
function deployer_batch_action_init_servers_ftp($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {
    _deployer_log("\n" . t('Initiating server connections (FTP).'), NULL, DEPLOYER_NO_DEBUG);
    // Load environment on which deploy is done.
    $env_field = field_get_items('node', $deploynode, 'field_environment_id');
    $deploy_context['env_id'] = $env_field[0]['target_id'];
    $env = node_load($deploy_context['env_id']);

    $server_instance = field_get_items('node', $env, 'field_server_instance');
    foreach ($server_instance as $server_field_collection) {
      $collection = entity_load('field_collection_item', array($server_field_collection['value']));

      // Set server properties.
      $hostname = $collection[$server_field_collection['value']]->field_hostname[LANGUAGE_NONE][0]['value'];
      $port = $collection[$server_field_collection['value']]->field_port[LANGUAGE_NONE][0]['value'];
      $username = $collection[$server_field_collection['value']]->field_username[LANGUAGE_NONE][0]['value'];
      $password = $collection[$server_field_collection['value']]->field_password[LANGUAGE_NONE][0]['value'];

      $webroot = $collection[$server_field_collection['value']]->field_path[LANGUAGE_NONE][0]['value'];

      // Remove trailing slash.
      $webroot = rtrim($webroot, "/");

      $owner_user = $collection[$server_field_collection['value']]->field_owner_user[LANGUAGE_NONE][0]['value'];
      $owner_group = $collection[$server_field_collection['value']]->field_owner_group[LANGUAGE_NONE][0]['value'];

      $active_mode = $collection[$server_field_collection['value']]->field_active_mode[LANGUAGE_NONE][0]['value'];
      if ($active_mode) {
        _deployer_log(t('Using active FTP mode'));
      }
      else {
        _deployer_log(t('Using passive FTP mode'));
      }

      // Define site instances.
      $site_instances = $collection[$server_field_collection['value']]->field_site_instance[LANGUAGE_NONE];
      $sites_urls = array();
      foreach ($site_instances as $site_instance) {
        $instance = entity_load('field_collection_item', array($site_instance['value']));
        $instance = $instance[$site_instance['value']];

        $site = $instance->field_site_dir[LANGUAGE_NONE][0]['value'];
        $url = $instance->field_site_url[LANGUAGE_NONE][0]['value'];
        $sites_urls[$site] = $url;
      }

      // Create the ftp object.
      _deployer_log(t('Setting up connection to %h% :', array('%h%' => $hostname)));
      $server_array = deployer_new_server_ftp($hostname, $port, $username, $password, $webroot, $active_mode);
      $server = $server_array[0];

      if (!$server->ftp) {
        $deploy_context['failed'] = 1;
        _deployer_log(t('    ==> failed. (CRITICAL)'));
        _deployer_log('    =>> message: ' . $server_array[1]);
        break;
      }
      else {
        _deployer_log(t('    => ok'));
      }

      $server->hostname = $hostname;
      $server->username = $username;
      $server->port = $port;
      $server->password = $password;
      $server->active_mode = $active_mode;

      $server->owner_user  = $owner_user;
      $server->owner_group  = $owner_group;
      $server->temp_archive = $server->temp_folder . '/.deploy_' . uniqid() . '.tar.gz';
      $server->ok = 1;
      $server->webroot_touched = 0;
      $server->database_touched = 0;
      $server->id = $username . '@' . $hostname . ':' . $webroot;
      $server->sites_urls = $sites_urls;
      $server->tag_str = '';

      $servers[] = $server;
    }
  }
}


/**
 * Transport webroot to server.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_transport_webroot_ftp($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed.
  if (!$deploy_context['failed']) {

    // Deploy.
    if ($deploy_context['mode'] == 'deploy') {
      _deployer_log("\n" . t('Starting transportation routine.'), NULL, DEPLOYER_NO_DEBUG);

      // Unpack archive.
      _deployer_log(t('Unpacking archive'));
      $identifier = uniqid();
      $temp_folder = file_directory_temp() . '/deploy/' . $identifier;
      drupal_mkdir($temp_folder, NULL, TRUE);

      exec('cd ' . $temp_folder . '; if [ $(pwd) = "' . $temp_folder . '" ]; then tar -xf ' . $deploy_context['archive'] . '; fi', $output, $return_stat);
      if ($return_stat !== 0) {
        _deployer_log(t('Cannot extract webroot archive!'));
        $deploy_context['failed'] = 1;
      }

      // Set tag string to version file in webroot.
      file_put_contents($temp_folder . '/.deployed_tag', $deploy_context['tag_str']);

      _deployer_log(t('Sending content to server'));
      foreach ($servers as &$server) {
        if (!$deploy_context['failed']) {
          $server->new_webroot = $server->webroot . '/' . uniqid() . '_webroot';

          // Create new folder, set rights.
          _deployer_ftp_cmd('ftp_mkdir', $server, array($server->new_webroot));
          _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $server->new_webroot));

          $contents = scandir($temp_folder);

          if (!empty($contents)) {
            foreach ($contents as $file) {
              // $file can also be a folder.
              if (!$deploy_context['failed']) {
                _deployer_ftp_put_recursive($server, $server->new_webroot, $temp_folder . '/' . $file, 1, 0, $deploy_context);
              }
            }
          }
          else {
            _deployer_log(t('Nothing to transport, archive empty'));
          }
        }
      }

      // Clean up.
      _deployer_log(t('Removing webroot archive from deployer-server'));
      exec('rm -rf ' . $temp_folder);
      unlink(drupal_realpath($deploy_context['archive']));
    }

    // Restore.
    elseif ($deploy_context['mode'] == 'restore') {
      _deployer_log("\n" . t('Starting transportation routine.'), NULL, DEPLOYER_NO_DEBUG);

      if ($deploy_context['restores']['webroot_backup'][0]) {
        // Unpack archive.
        _deployer_log(t('Unpacking archive'));
        $identifier = uniqid();
        $temp_folder = file_directory_temp() . '/deploy/' . $identifier;
        drupal_mkdir($temp_folder, NULL, TRUE);
        exec('cd ' . $temp_folder . '; if [ $(pwd) = "' . $temp_folder . '" ]; then tar -xf ' . drupal_realpath($deploy_context['restores']['webroot_backup'][0]) . '; fi', $output, $return_stat);
        if ($return_stat !== 0) {
          _deployer_log(t('Cannot extract webroot archive!'));
          $deploy_context['failed'] = 1;
        }
      }

      foreach ($servers as &$server) {

        if (!$deploy_context['failed']) {

          // Send webroot.
          if ($deploy_context['restores']['webroot_backup'][0]) {

            if (!$deploy_context['failed']) {
              $server->new_webroot = $server->webroot . '/' . uniqid() . '_webroot';

              // Create excludes array (full path) te exclude from
              // sending to server.
              $excludes = array();
              $contents = scandir($temp_folder);
              if (!empty($contents)) {
                _deployer_log(t('  Putting webroot'));
                foreach ($contents as $file) {

                  // Set special chmods array.
                  $chmod_777 = array();

                  // $file can also be a folder.
                  if (!$deploy_context['failed']) {

                    // No need to send whole webroot all the time!
                    // Lets exclude some stuff from transfer.
                    if ($file == 'sites' && is_dir($temp_folder . '/' . $file)) {
                      $sites_contents = scandir($temp_folder . '/' . $file);
                      foreach ($sites_contents as $site_folder) {
                        if ($temp_folder . '/' . $file . '/' . $site_folder != 'all' && is_dir($temp_folder . '/' . $file . '/' . $site_folder)) {
                          // Site folder found.
                          $settings_php = $temp_folder . '/' . $file . '/' . $site_folder . '/settings.php';
                          if (is_file($settings_php)) {
                            if (!$deploy_context['restore_settings_php']) {
                              $excludes[] = $settings_php;
                            }

                            // Transfer settings.php
                            // if no whole webroot restore is wanted.
                            elseif (!$deploy_context['restore_webroot']) {
                              _deployer_ftp_put_recursive($server, $server->new_webroot . '/' . $file . '/' . $site_folder, $settings_php, 1, 0, $deploy_context, $excludes);
                            }
                          }

                          $files_folder = $temp_folder . '/' . $file . '/' . $site_folder . '/files';
                          if (is_dir($files_folder)) {
                            if (!$deploy_context['restore_files']) {
                              $excludes[] = $files_folder;
                            }

                            // Transfer files folder
                            // if no whole webroot restore is wanted.
                            else {
                              if (!$deploy_context['restore_webroot']) {
                                _deployer_ftp_put_recursive($server, $server->new_webroot . '/' . $file . '/' . $site_folder, $files_folder, 1, 0, $deploy_context, $excludes);
                              }

                              $chmod_777[] = $server->new_webroot . '/' . $file . '/' . $site_folder . '/files';
                            }
                          }
                        }
                      }
                    }

                    // Send webroot with respect for excludes if needed.
                    if ($deploy_context['restore_webroot']) {
                      _deployer_ftp_put_recursive($server, $server->new_webroot, $temp_folder . '/' . $file, 1, 0, $deploy_context, $excludes);
                    }

                    // Special chmods.
                    if (isset($chmod_777)) {
                      foreach ($chmod_777 as $path) {
                        _deployer_log(t('     -> changing mode in %f to %m', array('%f' => $path, '%m' => 0777)));
                        _deployer_ftp_cmd('ftp_chmod', $server, array(0777, $path));
                        $files = _deployer_ftp_dirlist_recursive($server, $path);
                        foreach ($files as $file) {
                          _deployer_ftp_cmd('ftp_chmod', $server, array(0777, $path . '/' . $file));
                        }
                      }
                    }
                  }
                }
              }
            }

            // Create filelist from archive.
            $webroot_archive = new Archive_Tar(drupal_realpath($deploy_context['restores']['webroot_backup'][0]));
            $deploy_context['webroot_archive_content'] = $webroot_archive->listContent();
          }
        }
      }

      // Clean up.
      if (isset($temp_folder) && is_dir($temp_folder)) {
        _deployer_log(t('Removing webroot archive from deployer-server'));
        exec('rm -rf ' . $temp_folder);
      }

    }
  }
}

/**
 * Put RPC on server.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_put_rpc_ftp($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {
    _deployer_log("\n" . t('Transporting RPC to remote servers.'), NULL, DEPLOYER_NO_DEBUG);

    $deploy_context['rpc_key'] = _deployer_gen_uuid();
    $key_message = '<?php' . "\n";
    $key_message .= '$key = "' . $deploy_context['rpc_key'] . '";';

    $deploy_context['rpc_keyfile'] = 'private://deploy/' . uniqid() . '_keyfile.php';
    file_put_contents(drupal_realpath($deploy_context['rpc_keyfile']), $key_message);

    $module_path = DRUPAL_ROOT . '/' . drupal_get_path('module', 'deployer');
    foreach ($servers as &$server) {
      _deployer_ftp_put_file($server, $server->webroot . '/deploy_rpc.php', $module_path . '/mysql_ftp_snapshot/deploy_rpc.php', 1, $deploy_context);
      _deployer_ftp_put_file($server, $server->webroot . '/rpc_inc/mysqldump.php', $module_path . '/mysql_ftp_snapshot/rpc_inc/mysqldump.php', 1, $deploy_context);
      _deployer_ftp_put_file($server, $server->webroot . '/rpc_inc/mysql_load.php', $module_path . '/mysql_ftp_snapshot/rpc_inc/mysql_load.php', 1, $deploy_context);
      _deployer_ftp_put_file($server, $server->webroot . '/rpc_inc/.rpc_context.php', drupal_realpath($deploy_context['rpc_keyfile']), 1, $deploy_context);
    }

    _deployer_log(t('Done transporting RPC to remote servers.'));
  }
}

/**
 * Turn maintenance mode on.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_set_maintenance_on_ftp($deploynode, &$deploy_context, &$servers) {
  if (!$deploy_context['failed']) {
    _deployer_log("\n" . t('Setting sites on maintenance mode'));
    foreach ($servers as &$server) {
      foreach ($server->sites_urls as $site => $url) {
        if (!$deploy_context['failed']) {
          _deployer_rpc($url, 'action=maintenance_on', $deploy_context, 'warn');
        }
      }
    }
  }
}

/**
 * Turn maintenance mode off.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_set_maintenance_off_ftp($deploynode, &$deploy_context, &$servers) {
  _deployer_log("\n" . t('Turning off maintenance mode'));
  foreach ($servers as &$server) {
    foreach ($server->sites_urls as $site => $url) {
      if (!$deploy_context['failed']) {
        _deployer_rpc($url, 'action=maintenance_off', $deploy_context, 'warn');
      }
    }
  }
}

/**
 * Backup old webroot on server.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_backup_old_webroot_ftp($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {

    if ($deploy_context['webroot_backup']) {
      _deployer_log("\n" . t('Webroot backup routine.'), NULL, DEPLOYER_NO_DEBUG);

      foreach ($servers as &$server) {
        // Local backup folder.
        $server->local_webroot_backupfolder = 'private://deploy/backups/ftp_bak_' . uniqid();
        drupal_mkdir($server->local_webroot_backupfolder, NULL, TRUE);
        if (_deployer_ftp_cmd('ftp_chdir', $server, array($server->webroot))) {
          $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
          if (!empty($contents)) {
            foreach ($contents as $file) {
              if (!$deploy_context['failed']) {
                // Skip RPC files.
                if ($file != 'deploy_rpc.php' && $file != 'rpc_inc' && $file != basename($server->new_webroot)) {
                  // Download the folder structure.
                  _deployer_ftp_get_recursive($server, $file, $server->local_webroot_backupfolder, 1, 0, $deploy_context);
                }
              }
            }
          }

          // Set tagstring for backup.
          if (file_exists($server->local_webroot_backupfolder . '/.deployed_tag')) {
            $server->tag_str = file_get_contents($server->local_webroot_backupfolder . '/.deployed_tag');
          }

        }
        else {
          _deployer_log(t('Cannot change directory to @p', array('@p' => $server->webroot)));
          $deploy_context['failed'] = 1;
        }

        if ($deploy_context['failed']) {
          _deployer_log(t('Backup failed'));
        }
        else {
          $server->local_webroot_backupfolder_intact = 1;
        }
      }
      _deployer_log(t('Backup routine for webroot done.'));
    }
    else {
      _deployer_log(t('No backup needed.'));
    }
  }
}

/**
 * Backup old databases on server.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_backup_old_database_ftp($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {

    if ($deploy_context['database_backup']) {
      _deployer_log("\n" . t('Database backup routine'), NULL, DEPLOYER_NO_DEBUG);

      $server_counter = 0;
      foreach ($servers as &$server) {

        if ($server_counter == 0) {
          $server->db_backups = array();

          // Backup db's.
          $site_count = 0;
          foreach ($server->sites_urls as $site => $url) {
            if (!$deploy_context['failed']) {

              $server->db_backups[$site_count]['site'] = $site;
              $server->db_backups[$site_count]['localfile'] = 'private://deploy/backups/' . uniqid() . '_db.sql.gz';
              $server->db_backups[$site_count]['url'] = $url;

              _deployer_log(t('Creating database backup for site %s% on server %h%.', array('%s%' => $site, '%h%' => $server->id)));

              $response = _deployer_rpc($url, 'site=' . $site . '&action=dump', $deploy_context);
              if ($response) {
                if (!$saved_file = file_save_data($response->data, $server->db_backups[$site_count]['localfile'], FILE_EXISTS_REPLACE)) {
                  _deployer_log(t('Unable to save database backup on disk!'));
                  $deploy_context['failed'] = 1;
                }
                elseif (!filesize($server->db_backups[$site_count]['localfile'])) {
                  _deployer_log(t('Backupfile is empty!'));
                  $deploy_context['failed'] = 1;
                }
                else {
                  $size = filesize($server->db_backups[$site_count]['localfile']);
                  $size = format_size($size);
                  _deployer_log(t('Database saved to @f (@s)', array('@f' => $server->db_backups[$site_count]['localfile'], '@s' => $size)));
                }
              }

            }
            $site_count++;
          }
        }
        $server_counter++;
      }

    }
    else {
      _deployer_log(t('No database backup needed'));
    }
  }
}

/**
 * Unpack new webroot archive for deploy.
 *
 * Basically move it to the proper location on the ftp site.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_place_webroot_ftp($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed.
  if (!$deploy_context['failed']) {
    if ($deploy_context['mode'] == 'deploy' || ($deploy_context['mode'] == 'restore' && $deploy_context['restores']['webroot_backup'])) {
      _deployer_log("\n" . t('Switching new webroot with old webroot (actual deploy).'), NULL, DEPLOYER_NO_DEBUG);

      foreach ($servers as &$server) {
        // Mark temp old webroot as incomplete.
        // It's not created yet.
        $server->temp_old_webroot_intact = 0;

        // Place for old webroot:
        $server->temp_old_webroot = $server->webroot . '/' . uniqid() . '_old_webroot';

        // Place where webroot is build in case of restore:
        $server->temp_build = $server->webroot . '/' . uniqid() . '_build';

        if (!$deploy_context['failed']) {

          if (_deployer_ftp_cmd('ftp_chdir', $server, array($server->webroot))) {

            // Reset all filemodes in order to avoid faulty renames etc.
            _deployer_log(t('Resetting filemodes in order to avoid faulty renames etc due to missing rights.'));
            $filelist = _deployer_ftp_dirlist_recursive($server, $server->webroot);
            $files_count = count($filelist);
            _deployer_log(t('Counted @t files to change mode on.', array('@t' => $files_count)));

            // Track progress.
            $mod_count = 0;
            $mod_label = uniqid();
            _deployer_set_placeholder($mod_label, 0);
            _deployer_log(t('       Progress: !p %', array('!p' => $mod_label)));

            foreach ($filelist as $file) {
              $file = $server->webroot . '/' . $file;
              _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $file));

              // Track progress.
              $mod_count++;
              $percentage = ceil($mod_count / $files_count * 100);
              _deployer_set_placeholder($mod_label, $percentage);
            }

            // If mode is 'deploy',
            // move the current webroot to a temp folder,
            // and move the new webroot into place.
            // Then put files / settings.php back from old to live webroot.
            if ($deploy_context['mode'] == 'deploy') {
              // Move current content to $server->temp_old_webroot.
              _deployer_log(t('Moving content from webroot to @p:', array('@p' => $server->temp_old_webroot)));
              _deployer_ftp_cmd('ftp_mkdir', $server, array($server->temp_old_webroot));
              _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $server->temp_old_webroot));

              $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
              if (!empty($contents)) {
                foreach ($contents as $file) {

                  if (!$deploy_context['failed']) {
                    // Skip RPC files.
                    if ($file != '.' && $file != '..' && $file != 'deploy_rpc.php' && $file != 'rpc_inc' && $file != basename($server->new_webroot) && $file != basename($server->temp_old_webroot)) {
                      // Move content to temp old webroot folder.
                      /* _deployer_log(t('    => @f', array('@f' => $file)));*/
                      $status = _deployer_ftp_cmd('ftp_rename', $server, array($file, $server->temp_old_webroot . '/' . basename($file)));
                      if (!$status) {
                        $deploy_context['failed'] = 1;
                        _deployer_log(t('Moving file @f from current webroot to temp webroot failed! (CRITICAL)', array('@f' => $file)));
                      }
                    }
                  }
                }
                if (!$deploy_context['failed]']) {
                  // Mark temp old webroot as intact / complete.
                  // Makes it suitable for rollback.
                  $server->temp_old_webroot_intact = 1;
                }
              }
              if ($deploy_context['failed']) {
                _deployer_log(t('   => failed (CRITICAL)'));
              }
              else {
                _deployer_log(t('   => done'));
              }

              // Move new content to $server->webroot.
              _deployer_log(t('Moving content from new webroot to live webroot: '));
              if (_deployer_ftp_cmd('ftp_chdir', $server, array($server->new_webroot))) {
                $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
                _deployer_ftp_cmd('ftp_chdir', $server, array($server->webroot));
                if (!empty($contents)) {
                  foreach ($contents as $file) {

                    if (!$deploy_context['failed']) {
                      if ($file != '.' && $file != '..') {
                        // Move content to real webroot.
                        $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->new_webroot . '/' . $file, $server->webroot . '/' . basename($file)));
                        if (!$status) {
                          $deploy_context['failed'] = 1;
                          _deployer_log(t('Moving file @f from temp new webroot to real webroot failed! (CRITICAL)', array('@f' => $file)));
                        }
                      }
                    }
                  }
                }
              }
              else {
                _deployer_log(t('Cannot change directory to newly created webroot @p', array('@p' => $server->new_webroot)));
                $deploy_context['failed'] = 1;
              }
              if ($deploy_context['failed']) {
                _deployer_log(t('   => failed (CRITICAL)'));
              }
              else {
                _deployer_log(t('   => done'));
              }

              // Move files / settings.php from temp old webroot to
              // $server->webroot.
              $filelist = _deployer_ftp_dirlist_recursive($server, $server->temp_old_webroot . '/sites', $depth = 2);
              $server->temp_old_webroot_intact = 0;
              foreach ($filelist as $file) {
                $file = 'sites/' . $file;
                $move = 0;
                if (!$deploy_context['failed']) {
                  if (basename($file) == 'files') {
                    $move = 1;
                    _deployer_log(t('Moving files folder from temp_old webroot to new live webroot, and resetting filemodes.'));

                    // Set file modes recursive.
                    _deployer_ftp_cmd('ftp_chmod', $server, array(0777, $server->temp_old_webroot . '/' . $file));
                    $sub_files = _deployer_ftp_dirlist_recursive($server, $server->temp_old_webroot . '/' . $file);
                    foreach ($sub_files as $sub_file) {
                      _deployer_ftp_cmd('ftp_chmod', $server, array(0777, $server->temp_old_webroot . '/' . $file . '/' . $sub_file));
                    }
                  }
                  elseif (basename($file) == 'settings.php') {
                    $move = 1;
                    _deployer_log(t('Moving settings.php from temp_old webroot to new live webroot.'));
                  }

                  if ($move) {
                    // Old temp webroot not intact anymore.
                    $server->temp_old_webroot_intact = 0;

                    // Move content to real webroot.
                    _deployer_log(t('   > @f', array('@f' => $file)));
                    // Remove from new webroot. Shouldnt be in the repo anyway!
                    _deployer_ftp_delete_recursive($server, $server->webroot . '/' . $file, 0, 0);
                    _deployer_ftp_mksubdirs($server, '/', dirname($server->webroot . '/' . $file));
                    $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->temp_old_webroot . '/' . $file, $server->webroot . '/' . $file));
                    if (!$status) {
                      $deploy_context['failed'] = 1;
                      _deployer_log(t('   => failed! .(CRITICAL)'));
                    }
                  }
                }
              }
            }

            // Restore + restore webroot:
            // Upload to tmp_build.
            elseif ($deploy_context['mode'] == 'restore') {

              // Create build folder.
              $status = _deployer_ftp_cmd('ftp_mkdir', $server, array($server->temp_build));
              _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $server->temp_build));

              if (!$status) {
                _deployer_log(t('Unable to create build folder @f', array('@f' => $server->temp_build)));
              }
              else {
                _deployer_log(t('Created build folder on site'));
              }

              // Keep old webroot?
              if (!$deploy_context['restore_webroot']) {
                // Move current content to $server->temp_build.
                _deployer_log(t('Moving content from webroot to build folder @p: ', array('@p' => $server->temp_build)));
                _deployer_ftp_cmd('ftp_mkdir', $server, array($server->temp_build));
                _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $server->temp_build));

                $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
                if (!empty($contents)) {
                  foreach ($contents as $file) {

                    if (!$deploy_context['failed']) {
                      // Skip RPC files.
                      if ($file != '.' && $file != '..' && $file != 'deploy_rpc.php' && $file != 'rpc_inc' && $file != basename($server->new_webroot) && $file != basename($server->temp_build) && $file != basename($server->temp_build)) {
                        // Move content to temp old webroot folder.
                        /* _deployer_log(t('    => @f', array('@f' => $file)));*/
                        $status = _deployer_ftp_cmd('ftp_rename', $server, array($file, $server->temp_build . '/' . basename($file)));
                        if (!$status) {
                          $deploy_context['failed'] = 1;
                          _deployer_log(t('Moving file @f from current webroot to webroot build folder failed! (CRITICAL)', array('@f' => $file)));
                        }
                      }
                    }
                  }
                  if (!$deploy_context['failed]']) {
                    // Mark temp old webroot as intact / complete.
                    // Makes it suitable for rollback.
                    $server->temp_build_intact = 1;
                  }
                }
                if ($deploy_context['failed']) {
                  _deployer_log(t('   => failed (CRITICAL)'));
                }
                else {
                  _deployer_log(t('   => done'));
                }

              }

              // Restore webroot, discard old one.
              else {
                // Move current content to $server->temp_old_webroot.
                _deployer_log(t('Moving content from webroot to @p: ', array('@p' => $server->temp_old_webroot)));
                _deployer_ftp_cmd('ftp_mkdir', $server, array($server->temp_old_webroot));
                _deployer_ftp_cmd('ftp_chmod', $server, array(0755, $server->temp_old_webroot));

                $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
                if (!empty($contents)) {
                  foreach ($contents as $file) {

                    if (!$deploy_context['failed']) {
                      // Skip RPC files.
                      if ($file != '.' && $file != '..' && $file != 'deploy_rpc.php' && $file != 'rpc_inc' && $file != basename($server->new_webroot) && $file != basename($server->temp_old_webroot) && $file != basename($server->temp_build)) {
                        // Move content to temp old webroot folder.
                        /* _deployer_log(t('    => @f', array('@f' => $file)));*/
                        $status = _deployer_ftp_cmd('ftp_rename', $server, array($file, $server->temp_old_webroot . '/' . basename($file)));
                        if (!$status) {
                          $deploy_context['failed'] = 1;
                          _deployer_log(t('Moving file @f from current webroot to temp webroot failed! (CRITICAL)', array('@f' => $file)));
                        }
                      }
                    }
                  }
                  if (!$deploy_context['failed]']) {
                    // Mark temp old webroot as intact / complete.
                    // Makes it suitable for rollback.
                    $server->temp_old_webroot_intact = 1;
                  }
                }
                if ($deploy_context['failed']) {
                  _deployer_log(t('   => failed (CRITICAL)'));
                }
                else {
                  _deployer_log(t('   => done'));
                }

                // Move new content to $server->temp_build.
                _deployer_log(t('Moving content from webroot-to-restore to build folder.'));
                if (_deployer_ftp_cmd('ftp_chdir', $server, array($server->new_webroot))) {
                  $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
                  _deployer_ftp_cmd('ftp_chdir', $server, array($server->webroot));
                  if (!empty($contents)) {
                    foreach ($contents as $file) {

                      if (!$deploy_context['failed']) {
                        if ($file != '.' && $file != '..') {
                          // Move content to real webroot.
                          $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->new_webroot . '/' . $file, $server->temp_build . '/' . basename($file)));
                          if (!$status) {
                            $deploy_context['failed'] = 1;
                            _deployer_log(t('Moving file @f from temp new webroot to build folder @p failed! (CRITICAL)', array('@f' => $file, '@p' => $server->temp_build)));
                          }
                        }
                      }
                    }
                  }
                }
                else {
                  _deployer_log(t('Cannot change directory to newly created webroot @p', array('@p' => $server->new_webroot)));
                  $deploy_context['failed'] = 1;
                }
                if ($deploy_context['failed']) {
                  _deployer_log(t('   => failed (CRITICAL)'));
                }
                else {
                  _deployer_log(t('   => done'));
                }
              }

              // Move files from temp old webroot to
              // build folder.
              if (!$deploy_context['restore_files']) {

                // Old webroot is not transported to build folder.
                // Need to extract files folders out of old temp webroot.
                if ($deploy_context['restore_webroot']) {

                  // Check if path exists.
                  if (_deployer_ftp_cmd('ftp_size', $server, array($server->temp_old_webroot . '/sites'), FALSE) !== FALSE) {
                    $filelist = _deployer_ftp_dirlist_recursive($server, $server->temp_old_webroot . '/sites', $depth = 2);
                    foreach ($filelist as $file) {
                      $file = 'sites/' . $file;
                      $server->temp_old_webroot_intact = 0;
                      $move = 0;
                      if (!$deploy_context['failed']) {
                        if (basename($file) == 'files') {
                          $move = 1;
                          _deployer_log(t('Moving files folder from temp_old webroot to build folder.'));
                        }

                        if ($move) {
                          // Old temp webroot not intact anymore.
                          $server->temp_old_webroot_intact = 0;

                          // Move content to real webroot.
                          _deployer_log(t('   > @f', array('@f' => $file)));
                          // Remove from new webroot.
                          _deployer_ftp_delete_recursive($server, $server->temp_build . '/' . $file, 0, 0);
                          _deployer_ftp_mksubdirs($server, '/', dirname($server->temp_build . '/' . $file));
                          $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->temp_old_webroot . '/' . $file, $server->temp_build . '/' . $file));
                          if (!$status) {
                            $deploy_context['failed'] = 1;
                            _deployer_log(t('   => failed! .(CRITICAL)'));
                          }
                        }
                      }
                    }
                  }
                }
                else {
                  // Do nothing, old webroot was preserved,
                  // Including files folders.
                }
              }
              else {
                // Move new_webroot files to build folder.
                $filelist = _deployer_ftp_dirlist_recursive($server, $server->new_webroot . '/sites', $depth = 2);
                foreach ($filelist as $file) {
                  $file = 'sites/' . $file;
                  $move = 0;
                  if (!$deploy_context['failed']) {
                    if (basename($file) == 'files') {
                      $move = 1;
                      _deployer_log(t('Moving files folder from new webroot to build folder.'));
                    }

                    if ($move) {
                      // Move content to build folder.
                      _deployer_log(t('   > @f', array('@f' => $file)));
                      // Remove from new webroot.
                      _deployer_ftp_delete_recursive($server, $server->temp_build . '/' . $file, 0, 0);
                      _deployer_ftp_mksubdirs($server, '/', dirname($server->temp_build . '/' . $file));
                      $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->new_webroot . '/' . $file, $server->temp_build . '/' . $file));
                      if (!$status) {
                        $deploy_context['failed'] = 1;
                        _deployer_log(t('   => failed! .(CRITICAL)'));
                      }
                    }
                  }
                }
              }

              // Move settings.php from temp old webroot to
              // build folder.
              if (!$deploy_context['restore_settings_php']) {

                // Old webroot is not transported to build folder.
                // Need to extract settings.php's out of old temp webroot.
                if ($deploy_context['restore_webroot']) {
                  $filelist = _deployer_ftp_dirlist_recursive($server, $server->temp_old_webroot . '/sites', $depth = 2);
                  foreach ($filelist as $file) {
                    $file = 'sites/' . $file;
                    $move = 0;
                    if (!$deploy_context['failed']) {
                      if (basename($file) == 'settings.php') {
                        $move = 1;
                        _deployer_log(t('Moving settings.php from temp_old webroot to build folder.'));
                      }

                      if ($move) {
                        // Old temp webroot not intact anymore.
                        $server->temp_old_webroot_intact = 0;

                        // Move content to real webroot.
                        _deployer_log(t('   > @f', array('@f' => $file)));
                        // Remove from new webroot.
                        _deployer_ftp_delete_recursive($server, $server->temp_build . '/' . $file, 0, 0);
                        _deployer_ftp_mksubdirs($server, '/', dirname($server->temp_build . '/' . $file));
                        $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->temp_old_webroot . '/' . $file, $server->temp_build . '/' . $file));
                        if (!$status) {
                          $deploy_context['failed'] = 1;
                          _deployer_log(t('   => failed! .(CRITICAL)'));
                        }
                      }
                    }
                  }
                }
                else {
                  // Do nothing, old webroot was preserved,
                  // Including files folders.
                }
              }
              else {
                // Move new_webroot settings.php to build folder.
                $filelist = _deployer_ftp_dirlist_recursive($server, $server->new_webroot . '/sites', $depth = 2);
                foreach ($filelist as $file) {
                  $file = 'sites/' . $file;
                  $move = 0;
                  if (!$deploy_context['failed']) {
                    if (basename($file) == 'settings.php') {
                      $move = 1;
                      _deployer_log(t('Moving settings.php from new webroot to build folder.'));
                    }

                    if ($move) {
                      // Move content to build folder.
                      _deployer_log(t('   > @f', array('@f' => $file)));
                      // Remove from new webroot.
                      _deployer_ftp_delete_recursive($server, $server->temp_build . '/' . $file, 0, 0);
                      _deployer_ftp_mksubdirs($server, '/', dirname($server->temp_build . '/' . $file));
                      $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->new_webroot . '/' . $file, $server->temp_build . '/' . $file));
                      if (!$status) {
                        $deploy_context['failed'] = 1;
                        _deployer_log(t('   => failed! .(CRITICAL)'));
                      }
                    }
                  }
                }
              }

              // Move content in build folder to $server->webroot.
              _deployer_log(t('Moving content from build folder to @webroot', array('@webroot' => $server->webroot)));
              _deployer_ftp_cmd('ftp_chdir', $server, array($server->temp_build));
              $contents = _deployer_ftp_cmd('ftp_nlist', $server, array("-a ."));
              if (!empty($contents)) {
                foreach ($contents as $file) {

                  if ($file != '.' && $file != '..') {
                    if (!$deploy_context['failed']) {
                      // Move content to webroot.
                      $status = _deployer_ftp_cmd('ftp_rename', $server, array($file, $server->webroot . '/' . basename($file)));
                      if (!$status) {
                        $deploy_context['failed'] = 1;
                        _deployer_log(t('Moving file @f from build folder to webroot failed! (CRITICAL)', array('@f' => $file)));
                      }
                    }
                  }
                }
              }
              if ($deploy_context['failed']) {
                _deployer_log(t('   => failed (CRITICAL)'));
              }
              else {
                _deployer_log(t('   => done'));
              }

            }

          }
          else {
            _deployer_log(t('Cannot change directory to @p', array('@p' => $server->webroot)));
            $deploy_context['failed'] = 1;
          }

          $server->webroot_touched = 1;
        }
      }
    }
  }
}

/**
 * Load databases in case of restore.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_unpack_databases_ftp($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed.
  if (!$deploy_context['failed'] && $deploy_context['mode'] == 'restore') {
    if (isset($deploy_context['restores']['database_backup']) && !empty($deploy_context['restores']['database_backup'])) {

      _deployer_log("\n" . t('Restore routine for databases.'), NULL, DEPLOYER_NO_DEBUG);

      foreach ($servers as &$server) {
        if (!$deploy_context['failed']) {

          // Rollback db's.
          foreach ($deploy_context['restores']['database_backup'] as $backup) {

            if (!$deploy_context['failed']) {
              if (!isset($backup['site'])) {
                _deployer_log(t('Cannot rollback databases, no site or filename available! (CRITICAL)'));
                $deploy_context['failed'] = 1;
                continue;
              }

              if (isset($server->sites_urls[$backup['site']])) {

                // Mark database as touched.
                $server->database_touched = 1;

                $site = $backup['site'];
                $url = $server->sites_urls[$backup['site']];

                $identifier = uniqid();
                $temp_folder = file_directory_temp() . '/deploy/' . $identifier;
                drupal_mkdir($temp_folder, NULL, TRUE);
                exec('cd ' . $temp_folder . '; if [ $(pwd) = "' . $temp_folder . '" ]; then zcat ' . drupal_realpath($backup['localfile']) . ' > temp_db.sql; fi', $output, $return_stat);

                // Send file.
                $temp_remote_db = uniqid() . '.sql';
                $query = 'site=' . $site . '&action=restore' . '&file=' . $temp_remote_db;

                _deployer_ftp_put_file($server, $server->webroot . '/' . $temp_remote_db, $temp_folder . '/temp_db.sql', 1);
                $response = _deployer_rpc($url, $query, $deploy_context);
                _deployer_ftp_delete_recursive($server, $server->webroot . '/' . $temp_remote_db, 1, 0, $deploy_context, TRUE);

                if ($response) {
                  _deployer_log(t('Restore of db for site %s ok.', array('%s' => $backup['site'])));
                }
                else {
                  _deployer_log(t('Restore of db for site %s failed. (CRITICAL)', array('%s' => $backup['site'])));
                  $deploy_context['failed'] = 1;
                }

                // Clean up.
                exec('rm -rf ' . $temp_folder);
              }
              else {
                _deployer_log(t('No such site (%site) configured for this server instance. (CRITICAL)', array('%s' => $backup['site'])));
                $deploy_context['failed'] = 1;
              }
            }
          }
        }
      }
      _deployer_log(t('Restore routine for database done.'));
    }
  }
}

/**
 * Rollback webroot in case of failure.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_rollback_webroot_ftp($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed, otherwise no need for rollback.
  if ($deploy_context['failed']) {
    _deployer_log("\n" . t('Rollback routine for webroot.'), NULL, DEPLOYER_NO_DEBUG);

    foreach ($servers as &$server) {
      if ($deploy_context['webroot_backup']) {
        if ($server->webroot_touched) {
          if ($server->temp_old_webroot_intact || $server->local_webroot_backupfolder_intact) {
            // Get rid of content in webroot.
            $filelist = _deployer_ftp_dirlist_recursive($server, $server->webroot, $depth = 1);

            foreach ($filelist as $file) {
              if ($file != 'deploy_rpc.php' && $file != 'rpc_inc' && $file != basename($server->new_webroot) && $file != basename($server->temp_old_webroot)) {
                // Remove file / folder.
                _deployer_ftp_delete_recursive($server, $server->webroot . '/' . $file, 1, 0, $deploy_context, TRUE);
              }
            }
          }

          if ($server->temp_old_webroot_intact) {
            _deployer_log(t('Rolling back webroot from live folder on ftp site.'));
            $filelist = _deployer_ftp_dirlist_recursive($server, $server->temp_old_webroot, $depth = 1);
            foreach ($filelist as $file) {
              // Move every file / folder to real webroot.
              $status = _deployer_ftp_cmd('ftp_rename', $server, array($server->temp_old_webroot . '/' . $file, $server->webroot . '/' . $file));
              if (!$status) {
                _deployer_log(t('Moving @f from webroot backup to live webroot failed! (CRITICAL)', array('@f' => $file)));
              }
            }
          }
          elseif ($server->local_webroot_backupfolder_intact) {
            // Upload files webroot from local backup folder.
            _deployer_log(t('Rolling back webroot by uploading from complete backup.'));
            $contents = scandir(drupal_realpath($server->local_webroot_backupfolder));
            if (!empty($contents)) {
              foreach ($contents as $file) {
                _deployer_ftp_put_recursive($server, $server->webroot, drupal_realpath($server->local_webroot_backupfolder) . '/' . $file, 1, 0);
              }
            }
          }
          else {
            _deployer_log(t('CANNOT ROLLBACK, NO WEBROOT BACKUP SEEMS PRESENT! (CRITICAL)'));
          }
        }
        else {
          _deployer_log(t('Webroot not touched on server %h%, no need for rollback.', array('%h%' => $server->id)));
        }
      }
      else {
        _deployer_log(t('You choose not to do any backup, so no rollback possible.'));
      }
    }
  }
}

/**
 * Rollback databases in case of failure.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_rollback_database_ftp($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed, otherwise no need for rollback.
  if ($deploy_context['failed']) {
    _deployer_log("\n" . t('Rollback routine for database.'), NULL, DEPLOYER_NO_DEBUG);

    if ($deploy_context['database_backup']) {
      foreach ($servers as &$server) {

        if ($server->database_touched || $server->webroot_touched) {
          // Rollback db's.
          foreach ($server->db_backups as $backup) {

            if (!isset($backup['site']) || !isset($backup['localfile'])) {
              _deployer_log(t('Backup for this instance not present!! CANNOT ROLLBACK DATABASE! (CRITICAL)'));
              continue;
            }

            $identifier = uniqid();
            $temp_folder = file_directory_temp() . '/deploy/' . $identifier;
            drupal_mkdir($temp_folder, NULL, TRUE);
            exec('cd ' . $temp_folder . '; if [ $(pwd) = "' . $temp_folder . '" ]; then zcat ' . drupal_realpath($backup['localfile']) . ' > temp_db.sql; fi', $output, $return_stat);

            // Send file.
            $temp_remote_db = uniqid() . '.sql';
            $query = 'site=' . $backup['site'] . '&action=restore' . '&file=' . $temp_remote_db;

            _deployer_ftp_put_file($server, $server->webroot . '/' . $temp_remote_db, $temp_folder . '/temp_db.sql', 1);
            $response = _deployer_rpc($backup['url'], $query, $deploy_context);
            _deployer_ftp_delete_recursive($server, $server->webroot . '/' . $temp_remote_db, 1, 0, $deploy_context, TRUE);

            if ($response) {
              _deployer_log(t('Rollback of db for site %s ok.', array('%s' => $backup['site'])));
            }
            else {
              _deployer_log(t('Rollback of db for site %s failed!!! (CRITICAL)', array('%s' => $backup['site'])));
            }

            // Clean up.
            exec('rm -rf ' . $temp_folder);

          }
        }
        else {
          _deployer_log(t('Database and webroot not touched on server %h%, so no need to rollback.', array('%h%' => $server->id)));
        }
      }
    }
    else {
      _deployer_log(t('You choose not to make a database backup, so rollback is impossible.'));
    }
    _deployer_log(t('Rollback routine for database done.'));
  }
}


/**
 * Finish the deploy, clean up archives etc.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_finish_ftp($deploynode, &$deploy_context, &$servers) {

  _deployer_log("\n" . t('Finishing deploy now.'), NULL, DEPLOYER_NO_DEBUG);

  $backup_meta_data = array();

  // Prepare backup location on deployer server.
  $create_path = drupal_realpath('private://deploy/backups');
  if (!file_prepare_directory($create_path)) {
    drupal_mkdir('private://deploy/backups', NULL, TRUE);
  }

  // Keep track of failed backups.
  $backups_ok = -1;

  $server_counter = 0;
  foreach ($servers as &$server) {

    // Remove RPC.
    _deployer_ftp_delete_recursive($server, $server->webroot . '/rpc_inc', 1, 0, $deploy_context, TRUE);
    _deployer_ftp_delete_recursive($server, $server->webroot . '/deploy_rpc.php', 1, 0, $deploy_context, TRUE);

    if (isset($server->temp_old_webroot)) {
      _deployer_ftp_delete_recursive($server, $server->temp_old_webroot, 1, 0, $deploy_context, TRUE);
    }

    if (isset($server->new_webroot)) {
      _deployer_ftp_delete_recursive($server, $server->new_webroot, 1, 0, $deploy_context, TRUE);
    }

    if (isset($server->temp_build)) {
      _deployer_ftp_delete_recursive($server, $server->temp_build, 1, 0, $deploy_context, TRUE);
    }

    // Create webroot backup file.
    if (isset($server->local_webroot_backupfolder)) {

      // Only first server in cluster.
      if ($server_counter == 0) {
        $server->webroot_local_backupfile = 'private://deploy/backups/' . uniqid() . '_webroot.tar.gz';

        // Get tag string from backup, if present.
        if (file_exists($server->local_webroot_backupfolder . '/.deployed_tag')) {
          $server->tag_str = file_get_contents($server->local_webroot_backupfolder . '/.deployed_tag');
        }

        // Pack webroot backup into archive.
        // Remove backup folder.
        $webroot_tar = new Archive_Tar(drupal_realpath($server->webroot_local_backupfile));
        $filelist[0] = $server->local_webroot_backupfolder . '/';
        $webroot_tar->addModify($filelist, '.', $server->local_webroot_backupfolder . '/');
        file_unmanaged_delete_recursive($server->local_webroot_backupfolder);

        // Backup meta data.
        if (is_file(drupal_realpath($server->webroot_local_backupfile))) {
          $backup_meta_data[$server_counter]['webroot_backup'] = $server->webroot_local_backupfile;
          $backup_meta_data[$server_counter]['tag_str'] = $server->tag_str;
          // Create file entity.
          _deployer_create_save_file_entity($server->webroot_local_backupfile);
          if ($backups_ok != 0) {
            $backups_ok = 1;
          }
        }
        else {
          $backups_ok = 0;
        }
      }

    }

    // Check for database backupfiles.
    if (isset($server->db_backups[0]) && !empty($server->db_backups[0]['localfile'])) {

      $backup_counter = 0;
      foreach ($server->db_backups as $backup) {

        if (isset($backup['localfile'])) {
          // Only first server in cluster.
          if ($server_counter == 0) {

            // Backup meta data.
            if (is_file(drupal_realpath($backup['localfile']))) {
              $backup_meta_data[$server_counter]['database_backup'][$backup_counter]['site'] = $backup['site'];
              $backup_meta_data[$server_counter]['database_backup'][$backup_counter]['localfile'] = $backup['localfile'];

              // Create file entity.
              _deployer_create_save_file_entity($backup['localfile']);
              if ($backups_ok != 0) {
                $backups_ok = 1;
              }
            }
            else {
              $backups_ok = 0;
            }

            $server->db_backups[$backup_counter] = $backup;
          }

        }

        $backup_counter++;
      }
    }

    // Close connection.
    ftp_close($server->ftp);

    $server_counter++;
  }

  _deployer_log(t('Saving backup metadata.'));
  file_put_contents(drupal_realpath('private://deploy/backups/' . $deploynode->nid . '.meta'), drupal_json_encode($backup_meta_data));

  // Clean up.
  unlink(drupal_realpath($deploy_context['rpc_keyfile']));

  _deployer_log(t('Finished.'), NULL, DEPLOYER_NO_DEBUG);

  // Set deploystatus on node to 1 (finished) and save backup status.
  $temp_node = node_load($deploynode->nid);

  if ($deploy_context['failed']) {
    $temp_node->field_deploystatus[LANGUAGE_NONE][0]['value'] = 2;
  }
  else {
    $temp_node->field_deploystatus[LANGUAGE_NONE][0]['value'] = 1;
  }

  if ($backups_ok <= 0) {
    $backups_ok = 0;
  }
  $temp_node->field_backups_ok[LANGUAGE_NONE][0]['value'] = $backups_ok;

  node_save($temp_node);
}
