<?php
/**
 * @file
 * Functions used on ssh-connected environments during deploys.
 */

/**
 * Put files on remote server.
 *
 * @param object $server
 *   Net:SFTP object.
 * @param array $puts
 *   Array containing ('CRIT'|'WARN', [local file], [remote file]).
 *   If array[0] == 'CRIT', we will set $deploy_context['failed'] to 1.
 * @param array $deploy_context
 *   Deploy context.
 */
function _deployer_sftp_put(&$server, $puts, &$deploy_context = NULL) {
  foreach ($puts as $put) {
    $local = $put[1];
    $remote = $put[2];

    $size = format_size(filesize($local));
    _deployer_log(t('Transporting %f% (%size%) to %s%: ', array(
      '%f%' => basename($local),
      '%s%' => $server->hostname,
      '%size%' => $size,
    )));

    $transportation = $server->put($remote, $local, NET_SFTP_LOCAL_FILE);

    if ($put[0] == 'CRIT' && !empty($deploy_context) && !$transportation) {
      $deploy_context['failed'] = 1;
      _deployer_log(t('   ==> failed. (CRITICAL)'));
    }
    else {
      _deployer_log(t('   ==> ok'));
    }
  }
}

/**
 * Get files on remote server.
 *
 * @param object $server
 *   Net:SFTP object.
 * @param array $gets
 *   Array containing ('CRIT'|'WARN', [remote file], [local file]).
 *   If array[0] == 'CRIT', we will set $deploy_context['failed'] to 1.
 * @param array $deploy_context
 *   Deploy context.
 */
function _deployer_sftp_get(&$server, $gets, &$deploy_context = NULL) {
  foreach ($gets as $get) {
    $remote = $get[1];
    $local = $get[2];

    $size = format_size($server->size($remote));
    _deployer_log(t('Downloading %f% (%size%) from %s%: ', array(
      '%f%' => basename($remote),
      '%s%' => $server->hostname,
      '%size%' => $size,
    )));

    $transportation = $server->get($remote, $local);

    if (!$transportation || !filesize($local)) {
      _deployer_log(t('   ==> failed. (CRITICAL)'));
      if ($get[0] == 'CRIT' && !empty($deploy_context)) {
        $deploy_context['failed'] = 1;

      }
    }
    else {
      _deployer_log(t('   ==> ok'));
    }
  }
}

/**
 * Delete files on remote server.
 *
 * @param object $server
 *   Net:SFTP object.
 * @param array $dels
 *   Array containing ('CRIT'|'WARN', [remote file]).
 *   If array[0] == 'CRIT', we will set $deploy_context['failed'] to 1.
 * @param array $deploy_context
 *   Deploy context.
 */
function _deployer_sftp_delete_file(&$server, $dels, &$deploy_context = NULL) {
  foreach ($dels as $del) {
    $remote = $del[1];

    $size = format_size($server->size($remote));
    if ($size && !empty($size)) {
      _deployer_log(t('Deleting %f% (%size%) from %s%: ', array(
        '%f%' => basename($remote),
        '%s%' => $server->hostname,
        '%size%' => $size,
      )));

      $transportation = $server->delete($remote);

      if ($del[0] == 'CRIT' && !empty($deploy_context) && !$transportation) {
        $deploy_context['failed'] = 1;
        _deployer_log(t('   ==> failed. (CRITICAL)'));
      }
      else {
        _deployer_log(t('   ==> ok'));
      }
    }
  }
}

/**
 * Delete folders recursively on remote server.
 *
 * @param object $server
 *   Net:SFTP object.
 * @param array $dels
 *   Array containing ('CRIT'|'WARN', [remote file]).
 *   If array[0] == 'CRIT', we will set $deploy_context['failed'] to 1.
 * @param array $deploy_context
 *   Deploy context.
 */
function _deployer_sftp_delete_folder(&$server, $dels, &$deploy_context = NULL) {
  foreach ($dels as $del) {
    $remote = $del[1];

    _deployer_log(t('Deleting %f% from %s%: ', array(
      '%f%' => basename($remote),
      '%s%' => $server->hostname,
    )));

    $transportation = $server->delete($remote, TRUE);

    if ($del[0] == 'CRIT' && !empty($deploy_context) && !$transportation) {
      $deploy_context['failed'] = 1;
      _deployer_log(t('   ==> failed. (CRITICAL)'));
    }
    else {
      _deployer_log(t('   ==> ok'));
    }
  }
}

/**
 * Run shell commands on remote server.
 *
 * @param object $server
 *   Net:SFTP object.
 * @param array $commands
 *   Array containing command arrays,
 *   which contain (TRUE|FALSE, 'CRIT'|'WARN', [command], [working folder]).
 *   If array[0] == TRUE, perform command, even if prior commands failed.
 *   If array[1] == 'CRIT', we will set $deploy_context['failed'] to 1.
 * @param array $deploy_context
 *   Deploy context.
 *
 * @return array
 *   Contains command output in command order.
 */
function _deployer_ssh_command(&$server, $commands, &$deploy_context = NULL) {
  $failure = 0;
  $return = array();

  foreach ($commands as $key => $command) {
    $force = $command[0];
    $severity = $command[1];
    $cmd = $command[2];

    // Prevent password from being logged.
    $safe_cmd = preg_replace('/--password=[^ ]*/i', '--password=****', trim($cmd));

    if (!$failure || $force) {
      _deployer_log(t('Executing command { %cmd% }: ', array('%cmd%' => $safe_cmd)));

      if (isset($command[3])) {
        $folder = $command[3];
        $output = trim($server->exec('cd ' . $folder . '; if [ $(pwd) = "' . $folder . '" ]; then ' . $cmd . '; fi'));
      }
      else {
        $output = trim($server->exec($cmd));
      }
      $return[$key] = $output;

      if ($server->getExitStatus() > 0) {
        if ($severity == 'CRIT') {
          // Command failed.
          $failure = 1;

          // Critical, set context to failed.
          $deploy_context['failed'] = 1;
          _deployer_log(t('    => failed! (CRITICAL) (%e)', array('%e' => $server->getExitStatus())));

          // Print error output.
          _deployer_log('       ' . t('Output:'));
          _deployer_log($output);
        }
        else {
          // Asume $cmd[2] == 'WARN'.
          // Warning.
          _deployer_log(t('    => failed! (WARNING)'));
        }
      }
      else {
        _deployer_log(t('    => ok'));
      }
    }
  }
  return $return;
}

/**
 * Build server objects.
 *
 * @param object $deploynode
 *   Node object.
 * @param array $deploy_context
 *   Context array.
 * @param array $servers
 *   Servers array containing server objects.
 */
function deployer_batch_action_init_servers_ssh($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {
    _deployer_log("\n" . t('Initiating server connections (SSH).'), NULL, DEPLOYER_NO_DEBUG);
    // Load environment on which deploy is done.
    $env = $deploy_context['environment'];

    $server_instance = field_get_items('node', $env, 'field_server_instance');
    foreach ($server_instance as $server_field_collection) {
      $collection = entity_load('field_collection_item', array($server_field_collection['value']));

      $hostname = $collection[$server_field_collection['value']]->field_hostname[LANGUAGE_NONE][0]['value'];
      $port = $collection[$server_field_collection['value']]->field_port[LANGUAGE_NONE][0]['value'];
      $username = $collection[$server_field_collection['value']]->field_username[LANGUAGE_NONE][0]['value'];
      $webroot = $collection[$server_field_collection['value']]->field_path[LANGUAGE_NONE][0]['value'];

      $tmpfolder = '';
      if (isset($collection[$server_field_collection['value']]->field_temp_folder[LANGUAGE_NONE][0]['value'])) {
        $tmpfolder = $collection[$server_field_collection['value']]->field_temp_folder[LANGUAGE_NONE][0]['value'];
      }

      if (empty($tmpfolder)) {
        $tmpfolder = '/tmp/';
      }

      // Remove trailing slash.
      $webroot = rtrim($webroot, "/");

      $owner_user = $collection[$server_field_collection['value']]->field_owner_user[LANGUAGE_NONE][0]['value'];
      $owner_group = $collection[$server_field_collection['value']]->field_owner_group[LANGUAGE_NONE][0]['value'];

      // Create the ssh and sftp object.
      // Actually returns SFTP object, which is extended SSH2 object.
      _deployer_log(t('Setting up connection to %h% :', array('%h%' => $hostname)));
      $ssh_server_array = deployer_new_server_ssh($hostname, $port, $username, $webroot);
      $server = $ssh_server_array[0];

      if (!$server) {
        $deploy_context['failed'] = 1;
        _deployer_log(t('    ==> failed. (CRITICAL)'));
        _deployer_log('    =>> message: ' . $ssh_server_array[1]);
        break;
      }
      else {
        _deployer_log(t('    => ok'));
      }

      $server->owner_user  = $owner_user;
      $server->owner_group  = $owner_group;
      $server->temp_archive = $tmpfolder . '.deploy_' . uniqid() . '.tar.gz';
      $server->temp_folder = rtrim($tmpfolder, '/');
      $server->ok = 1;
      $server->webroot_touched = 0;
      $server->database_touched = 0;
      $server->id = $username . '@' . $hostname . ':' . $webroot;

      // Set drush default.
      // This is changed later if there's a valid drush exec.
      $server->drush = 'drush';

      // Create nested sftp object for ssh server object.
      $sftp_array = deployer_new_server_ssh($hostname, $port, $username, $webroot);
      $server->sftp = $sftp_array[0];

      if (!$server->sftp) {
        $deploy_context['failed'] = 1;
        _deployer_log(t('Faled to create sftp tunnel. (CRITICAL)'));
        _deployer_log('    =>> message: ' . $sftp_array[1]);
        break;
      }
      $server->sftp->hostname = $hostname;

      // Get current drupal version.
      $drupal_version = _deployer_get_drupal_version_ssh($server);
      if (!empty($drupal_version) && $drupal_version != 'DRUSH ERROR') {
        $server->current_drupal_version = $drupal_version;
      }
      else {
        $server->current_drupal_version = FALSE;
      }

      $servers[] = $server;
    }
  }
}

/**
 * Estimate if server has enough free space.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_estimate_free_space_ssh($deploynode, &$deploy_context, &$servers) {
  if (!$deploy_context['failed']) {
    $server_counter = 0;
    foreach ($servers as &$server) {

      _deployer_log("\n" . t('Checking temp folder on %h% : ', array('%h%' => $server->id)));
      // Get webroot space and free disk space.
      _deployer_ssh_command($server, array(
        'temp_check' => array(
          FALSE,
          'CRIT',
          '[ ! -d "' . $server->temp_folder . '" ] && mkdir -p "' . $server->temp_folder . '"; [ -d "' . $server->temp_folder . '" ]',
          $server->webroot,
        ),
      ), $deploy_context);

      // Stop if temp folder is not ok.
      if (!$deploy_context['failed']) {
        return FALSE;
      }

      _deployer_log("\n" . t('Estimating disk space on %h% : ', array('%h%' => $server->id)));
      // Get webroot space and free disk space.
      $output = _deployer_ssh_command($server, array(
        'webroot_size' => array(
          FALSE,
          'WARN',
          'du -bs | awk \'{print $1}\'',
          $server->webroot,
        ),
        'free_space' => array(
          FALSE,
          'WARN',
          'df -B1 ' . $server->temp_folder . ' | tail -n 1 | awk \'{print $4}\'',
          $server->webroot,
        ),
      ), $deploy_context);

      $restore_size = 0;
      $free_disk_space = trim($output['free_space']);
      $webroot_size = trim($output['webroot_size']);
      if (is_numeric($free_disk_space) && is_numeric($webroot_size)) {
        // Free space needs to be higher if we need a full backup.
        $size_factor = 0.2;
        if ($deploy_context['mode'] == 'deploy') {
          // Needed for packing files folders etc and restoring after deploy.
          $size_factor += 1;
        }
        if ($deploy_context['webroot_backup']) {
          $size_factor += 1;
        }

        if ($deploy_context['mode'] == 'restore' && isset($deploy_context['restores']['webroot_backup'][$server_counter])) {
          $size = format_size(filesize(drupal_realpath($deploy_context['restores']['webroot_backup'][$server_counter])));
          // Assume tarball will expand to 2 times it's size.
          $restore_size = ($size * 2);
        }

        $needed_space = ($webroot_size * $size_factor) + $restore_size;
        if (($free_disk_space - $needed_space) > 0) {
          _deployer_log(t('Enough free disk space available: !space', array('!space' => format_size($free_disk_space))));
          _deployer_log(t('Webroot of size !space, including backups must fit', array('!space' => format_size($webroot_size))));
        }
        else {
          $deploy_context['failed'] = 1;
          _deployer_log(t('Not enough disk space available: !space  (CRITICAL)', array('!space' => format_size($free_disk_space))));
          _deployer_log(t('Cannot fit webroot of size !space and/or restore files, including backups (CRITICAL)', array('!space' => format_size($webroot_size))));
          _deployer_log(t('...Or I just dont like you MUHAHA...'));
        }
      }
      else {
        _deployer_log(t('Warning: Unable to read webroot size or free disk space.'));
      }

      $server_counter++;
    }
  }
}

/**
 * Set drush version to use, if any provided.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_set_transport_drush_ssh($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {

    _deployer_log("\n" . t('Determining drush version to use.'));
    $env = $deploy_context['environment'];
    $drush_field = field_get_items('node', $env, 'field_drush_version');

    if (isset($drush_field[0]['value']) && !empty($drush_field[0]['value'])) {
      $drush_version = $drush_field[0]['value'];
      if ($drush_version != 'remote') {
        $deploy_context['drush'] = _deployer_get_drush($drush_version);
        if ($deploy_context['drush']) {

          // Pack and transport drush.
          _deployer_log(t('Drush version: %d%', array('%d%' => basename($deploy_context['drush']))));

          // Create drush tarball.
          if (is_dir($deploy_context['drush'])) {
            _deployer_log(t('Packing drush'));
            $deploy_context['drush_archive'] = 'private://' . uniqid() . '_drush.tar.gz';
            $drush_tar = new Archive_Tar(drupal_realpath($deploy_context['drush_archive']));

            // Check / patch drush.
            _deployer_check_patch_drush($deploy_context['drush']);

            $filelist[0] = $deploy_context['drush'] . '/';
            if ($drush_tar->addModify($filelist, '.', $deploy_context['drush'] . '/')) {
              // Transport drush to each server.
              _deployer_log(t('Sending drush archive to remote host.'));
              foreach ($servers as &$server) {
                if (!$deploy_context['failed']) {
                  $server->drush_folder = $server->temp_folder . '/.deploy_drush_' . uniqid();
                  $server->drush_archive = $server->drush_folder . '.tar.gz';

                  // Create temp dir on remote server.
                  _deployer_ssh_command($server->sftp, array(
                    array(
                      FALSE,
                      'CRIT',
                      'mkdir -p ' .  $server->drush_folder,
                      $server->temp_folder,
                    ),
                  ), $deploy_context);

                  // Send to server.
                  _deployer_sftp_put($server->sftp, array(
                    array(
                      'CRIT',
                      drupal_realpath($deploy_context['drush_archive']),
                      $server->drush_archive,
                    ),
                  ), $deploy_context);

                  // Unpack drush on server.
                  if (!$deploy_context['failed']) {
                    _deployer_ssh_command($server->sftp, array(
                      array(
                        FALSE,
                        'CRIT',
                        'tar -xf ' .  $server->drush_archive,
                        $server->drush_folder,
                      ),
                    ), $deploy_context);
                  }

                  if (!$deploy_context['failed']) {
                    $server->drush = $server->drush_folder . '/drush';
                  }

                }
              }

              // Remove drush archive from deployer server.
              _deployer_log(t('Removing drush archive from deployer-server'));
              unlink(drupal_realpath($deploy_context['drush_archive']));

            }
            else {
              // Use remote drush.
              _deployer_log(t('Packing drush failed. Falling back to remote drush.'));
              unset($deploy_context['drush']);
              unset($deploy_context['drush_archive']);
            }
          }
          else {
            _deployer_log(t('Drush folder not available. Falling back to remote drush.'));
            unset($deploy_context['drush']);
          }
        }
      }
    }
  }
}

/**
 * Send RPC.
 *
 * Used for parsing settings.php in case of restore.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_transport_rpc_ssh($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {
    if ($deploy_context['mode'] == 'restore') {
      _deployer_log(t('Sending RPC for DB restore to remote host.'));
      foreach ($servers as &$server) {
        if (!$deploy_context['failed']) {
          $server->rpc = $server->temp_folder . '/.' . uniqid() . '.rpc_parse_settings.php';
          // Send to server.
          _deployer_sftp_put($server->sftp, array(
            array(
              'CRIT',
              drupal_get_path('module', 'deployer') . '/includes/CLI_parse_settings.php',
              $server->rpc,
            ),
          ), $deploy_context);
        }
      }
    }
  }
}

/**
 * Transport webroot archive to server.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_transport_archive_ssh($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed.
  if (!$deploy_context['failed']) {

    // Deploy.
    if ($deploy_context['mode'] == 'deploy') {
      _deployer_log("\n" . t('Starting transportation routine.'), NULL, DEPLOYER_NO_DEBUG);
      foreach ($servers as &$server) {

        if (!$deploy_context['failed']) {

          // Send webroot archive to server.
          _deployer_sftp_put($server->sftp, array(
            array(
              'CRIT',
              drupal_realpath($deploy_context['archive']),
              $server->temp_archive,
            ),
          ), $deploy_context);

          // Check remote archive.
          _deployer_ssh_command($server, array(
            array(
              FALSE,
              'CRIT',
              'tar -tf ' . $server->temp_archive,
              NULL,
            ),
          ), $deploy_context);
        }
      }

      // Clean up.
      _deployer_log(t('Removing webroot archive from deployer-server'));
      unlink(drupal_realpath($deploy_context['archive']));
    }

    // Restore.
    elseif ($deploy_context['mode'] == 'restore') {
      _deployer_log("\n" . t('Starting transportation routine.'), NULL, DEPLOYER_NO_DEBUG);

      $server_counter = 0;
      foreach ($servers as &$server) {

        if (!$deploy_context['failed']) {

          // Send webroot archive to server.
          // Create filelist from archive to use later.
          if ($deploy_context['restores']['webroot_backup'][$server_counter]) {
            _deployer_sftp_put($server->sftp, array(
              array(
                'CRIT',
                drupal_realpath($deploy_context['restores']['webroot_backup'][$server_counter]),
                $server->temp_archive,
              ),
            ), $deploy_context);

            // Check remote archive.
            _deployer_ssh_command($server, array(
              array(
                FALSE,
                'CRIT',
                'tar -tf ' . $server->temp_archive,
                NULL,
              ),
            ), $deploy_context);

            // Create filelist from archive.
            $webroot_archive = new Archive_Tar(drupal_realpath($deploy_context['restores']['webroot_backup'][$server_counter]));
            $deploy_context['webroot_archive_content'] = $webroot_archive->listContent();
          }

          // Send database archive to server.
          if (isset($deploy_context['restores']['database_backup']) && !empty($deploy_context['restores']['database_backup'])) {

            $puts = array();
            $commands = array();
            foreach ($deploy_context['restores']['database_backup'] as $backup) {

              // Send file to remote host.
              $puts[] = array(
                'CRIT',
                drupal_realpath($backup['localfile']),
                $server->temp_db_archive[$backup['site']],
              );

              // Check file integrity.
              $commands[] = array(
                FALSE,
                'CRIT',
                'gunzip -t ' . $server->temp_db_archive[$backup['site']],
                NULL,
              );
            }
            _deployer_sftp_put($server->sftp, $puts, $deploy_context);
            _deployer_ssh_command($server, $commands, $deploy_context);
          }
        }
        // $server_counter++;
        // Only use first backup!
      }
    }
  }
}

/**
 * Backup old webroot on server.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_backup_old_webroot_ssh($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {

    if ($deploy_context['webroot_backup']) {
      _deployer_log("\n" . t('Webroot backup routine.'), NULL, DEPLOYER_NO_DEBUG);

      $server_count = 0;
      foreach ($servers as &$server) {

        // Only on first server in cluster, of mode is backup.
        if ($deploy_context['mode'] != 'backup' || $server_count == 0) {

          $server->webroot_backupfile = $server->temp_folder . '/.deploy_backup_webroot_' . uniqid() . '.tar.gz';
          _deployer_log(t('Creating backup on %h%', array('%h%' => $server->id)));

          // Create remote archive.
          $output = _deployer_ssh_command($server, array(
            // Check at first if we have write permissions for all files!
            0 => array(
              FALSE,
              'CRIT',
              'noperm=$(find . ! -writable); [ "$noperm" != "" ] && echo "No permissions for some files" && exit 1 || exit 0',
              $server->webroot,
            ),
            // Set mode to read only to make sure files are freezed.
            1 => array(
              FALSE,
              'CRIT',
              'find . -mindepth 1 -maxdepth 1 -exec chmod -R 444 "{}" \;',
              $server->webroot,
            ),
            2 => array(
              FALSE,
              'CRIT',
              'tar -cpzf ' . $server->webroot_backupfile . ' .',
              $server->webroot,
            ),
            // Set mode back to readable after backup.
            3 => array(
              TRUE,
              'CRIT',
              'find . -mindepth 1 -maxdepth 1 -exec chown -R ' . $server->owner_user . ':' . $server->owner_group . ' "{}" \;',
              $server->webroot,
            ),
            4 => array(
              TRUE,
              'CRIT',
              'find . -mindepth 1 -maxdepth 1 -exec chmod -R 750 "{}" \;',
              $server->webroot,
            ),
            // Set mode on drupal files folder.
            5 => array(
              TRUE,
              'CRIT',
              'find sites -mindepth 2 -maxdepth 2 -type d -name files -exec chmod -R 770 "{}" \; -exec chmod 7770 "{}" \;',
              $server->webroot,
            ),
            // Test archive.
            6 => array(
              FALSE,
              'CRIT',
              'tar -tf ' . $server->webroot_backupfile,
              $server->webroot,
            ),
            'tag_str' => array(
              FALSE,
              'WARN',
              '[ -f .deployed_tag ] && cat .deployed_tag || exit 0',
              $server->webroot,
            ),
          ), $deploy_context);

          $server->tag_str = $output['tag_str'];

          if ($deploy_context['failed']) {
            _deployer_log(t('Backup failed! (CRITICAL)'));
            if (!empty($output[0])) {
              _deployer_log($output[0]);
            }
          }
        }

        $server_count++;
      }
      _deployer_log(t('Backup routine for webroot done.'));
    }
    else {
      _deployer_log(t('No backup needed.'));
    }
  }
}

/**
 * Read sites aliases with drush.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_read_site_aliases_ssh($deploynode, &$deploy_context, &$servers) {
  if (!$deploy_context['failed']) {
    foreach ($servers as &$server) {
      _deployer_log("\n" . t('Reading drupal sites on %h% : ', array('%h%' => $server->id)));

      // Get drupal sites with drush.
      $output = _deployer_ssh_command($server, array(
        'sites' => array(
          FALSE,
          'CRIT',
          "find sites/ -mindepth 1 -maxdepth 1 -type d | sed 's|^sites/||g' | grep -v '^all$'",
          $server->webroot,
        ),
      ), $deploy_context);

      // Assume site-name on each newline.
      $server->sites = explode("\n", $output['sites']);

      if (!$deploy_context['failed']) {
        if (count($server->sites) == 0 || strlen($server->sites[0]) == 0) {
          $server->sites[0] = 'default';
          $deploy_context['failed'] = 0;
        }
      }
      else {
        _deployer_log(t('Reading sites failed. Assuming site "default"'));
        $server->sites = array('default');
        $deploy_context['failed'] = 0;
      }

      _deployer_log(t('Sites found: %s%', array('%s%' => $output['sites'])));
    }
  }
}

/**
 * Backup old databases on server.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_backup_old_database_ssh($deploynode, &$deploy_context, &$servers) {
  // Check if routine failed.
  if (!$deploy_context['failed']) {

    if ($deploy_context['database_backup']) {
      _deployer_log("\n" . t('Database backup routine'), NULL, DEPLOYER_NO_DEBUG);

      $server_count = 0;
      foreach ($servers as &$server) {

        // Only on first server in cluster, of mode is backup.
        if ($deploy_context['mode'] != 'backup' || $server_count == 0) {

          $server->db_backups = array();

          // Backup db's.
          $site_count = 0;
          foreach ($server->sites as $site) {

            $site = preg_replace('/.*#/i', '', trim($site));
            $server->db_backups[$site_count]['site'] = $site;
            $server->db_backups[$site_count]['filename'] = $server->temp_folder . '/.deploy_backup_' . $site . '_sql.gz';

            _deployer_log(t('Creating database backup for site %s% on server %h%.', array('%s%' => $site, '%h%' => $server->id)));

            if ($site == 'default') {
              $output = _deployer_ssh_command($server, array(
                // Get sql connection string.
                array(
                  FALSE,
                  'CRIT',
                  $server->drush . ' sql-connect 2>/dev/null',
                  $server->webroot,
                ),
                // Dump default site.
                array(
                  FALSE,
                  'CRIT',
                  $server->drush . ' sql-dump | gzip > ' . $server->db_backups[$site_count]['filename'],
                  $server->webroot,
                ),
                array(
                  FALSE,
                  'CRIT',
                  'gunzip -t ' . $server->db_backups[$site_count]['filename'],
                  $server->webroot,
                ),
              ), $deploy_context);
            }
            else {
              $output = _deployer_ssh_command($server, array(
                // Get sql connection string.
                array(
                  FALSE,
                  'CRIT',
                  $server->drush . ' --uri=' . $site .  ' sql-connect 2>/dev/null',
                  $server->webroot,
                ),
                // Dump site.
                array(
                  FALSE,
                  'CRIT',
                  $server->drush . ' --uri=' . $site .  ' sql-dump | gzip > ' . $server->db_backups[$site_count]['filename'],
                  $server->webroot,
                ),
                // Test archive.
                array(
                  FALSE,
                  'CRIT',
                  'gunzip -t ' . $server->db_backups[$site_count]['filename'],
                  $server->webroot,
                ),
              ), $deploy_context);
            }

            if (!empty($output[0])) {
              $server->db_backups[$site_count]['sql-connect'] = $output[0];
            }
            else {
              $deploy_context['failed'] = 1;
              _deployer_log(t('Unable to capture sql connection string'));
            }

            if (stristr($output[1], 'system()') || stristr($output[1], 'Got error')) {
              _deployer_log(t('Drush sql dump failed. (CRITICAL)'));
              $deploy_context['failed'] = 1;
            }

            if ($deploy_context['failed']) {
              $dels = array();
              $dels[] = array('WARN', $server->db_backups[$site_count]['filename']);
              _deployer_sftp_delete_file($server->sftp, $dels);
              unset($server->db_backups[$site_count]['filename']);
            }

            $site_count++;
          }
        }

        $server_count++;
      }
    }
    else {
      _deployer_log(t('No database backup needed'));
    }
  }
}

/**
 * Unpack new webroot archive for deploy.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_unpack_webroot_ssh($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed.
  if (!$deploy_context['failed']) {
    if ($deploy_context['mode'] == 'deploy' || ($deploy_context['mode'] == 'restore' && $deploy_context['restores']['webroot_backup'])) {
      _deployer_log("\n" . t('Unpacking archives (actual deploy).'), NULL, DEPLOYER_NO_DEBUG);

      foreach ($servers as &$server) {
        if (!$deploy_context['failed']) {

          $temp_files_archive = $server->temp_folder . '/' . uniqid() . '.tar.gz';

          $commands = array();
          $std_commands = array();

          /*Set standard deploy commands.*/

          // Backup files / settings.php
          $std_commands[] = array(
            FALSE,
            'CRIT',
            'tar -cpf ' . $temp_files_archive . ' $(find sites -mindepth 2 -maxdepth 2 -type f -name settings.php -o -type d -name files)',
            $server->webroot,
          );

          // PREPARE: Set mode on webroot.
          // Otherwise sometimes we cannot deploy as unprivileged user.
          $commands[] = array(
            FALSE,
            'CRIT',
            'find . -mindepth 1 -maxdepth 1 -exec chmod -R 750 "{}" \;',
            $server->webroot,
          );
          // Set mode on drupal files folder.
          $commands[] = array(
            FALSE,
            'CRIT',
            'find sites -mindepth 2 -maxdepth 2 -type d -name files -exec chmod -R 770 "{}" \; -exec chmod 7770 "{}" \;',
            $server->webroot,
          );

          // Remove files, including hidden files.
          $std_commands[] = array(
            FALSE,
            'CRIT',
            'find . -mindepth 1 -maxdepth 1 -exec rm -rf "{}" \;',
            $server->webroot,
          );

          // Extract new webroot archive.
          $std_commands[] = array(
            FALSE,
            'CRIT',
            'tar -xf ' . $server->temp_archive,
            $server->webroot,
          );

          // Remove sites/*/files and sites/*/settings.php from archive.
          // This should not happen, but it's possible.
          $std_commands[] = array(
            FALSE,
            'CRIT',
            'find sites -mindepth 2 -maxdepth 2 -type d -name files -exec rm -rf "{}" \;',
            $server->webroot,
          );
          $std_commands[] = array(
            FALSE,
            'CRIT',
            'find sites -mindepth 2 -maxdepth 2 -type f -name settings.php -exec rm -f "{}" \;',
            $server->webroot,
          );

          // Extract files / settings.php.
          $std_commands[] = array(
            FALSE,
            'CRIT',
            'tar -xf ' . $temp_files_archive . ' --overwrite --no-overwrite-dir',
            $server->webroot,
          );

          // Deploy commands.
          if ($deploy_context['mode'] == 'deploy') {
            $commands = $std_commands;
          }

          // Restore commands.
          elseif ($deploy_context['mode'] == 'restore') {

            // Whole webroot.
            if ($deploy_context['restore_webroot'] && $deploy_context['restore_files'] && $deploy_context['restore_settings_php']) {
              // Remove files, including hidden files.
              $commands[] = array(
                FALSE,
                'CRIT',
                'find . -mindepth 1 -maxdepth 1 -exec rm -rf "{}" \;',
                $server->webroot,
              );
              // Extract files / settings.php.
              $commands[] = array(
                FALSE,
                'CRIT',
                'tar -xf ' . $server->temp_archive,
                $server->webroot,
              );
            }

            else {
              // Settings.php.
              if ($deploy_context['restore_settings_php']) {
                // Remove old settings.php.
                $commands[] = array(
                  FALSE,
                  'CRIT',
                  'find sites -mindepth 2 -maxdepth 2 -type f -name settings.php -exec rm -f "{}" \;',
                  $server->webroot,
                );

                // Extract only settings.php in sites/*/ folder, not any deeper.
                $extract = '';
                foreach ($deploy_context['webroot_archive_content'] as $file) {
                  if (basename($file['filename']) == 'settings.php') {
                    $fileparts = explode('/', $file['filename']);
                    if (($fileparts[0] == '.' && $fileparts[1] == 'sites' && $fileparts[3] == 'settings.php') || ($fileparts[0] == 'sites' && $fileparts[2] == 'settings.php')) {
                      $extract .= ' -x ' . $file['filename'];
                    }
                  }
                }

                // Extract.
                if (!empty($extract)) {
                  $commands[] = array(
                    FALSE,
                    'CRIT',
                    'tar' . $extract . ' -f ' . $server->temp_archive,
                    $server->webroot,
                  );
                }
                else {
                  _deployer_log(t('No settings.php found in archive on right level.'));
                  $deploy_context['failed'] = 1;
                }
              }

              // Files folder.
              if ($deploy_context['restore_files']) {
                // Remove old files.
                $commands[] = array(
                  FALSE,
                  'CRIT',
                  'find sites -mindepth 2 -maxdepth 2 -type d -name files -exec rm -rf "{}" \;',
                  $server->webroot,
                );

                // Extract only files in sites/*/ folder, not any deeper.
                $extract = '';
                foreach ($deploy_context['webroot_archive_content'] as $file) {
                  if (basename($file['filename']) == 'files') {
                    $fileparts = explode('/', $file['filename']);
                    if ($fileparts[0] == '.' && $fileparts[1] == 'sites' && $fileparts[3] == 'files' && !isset($fileparts[5])) {
                      $extract .= ' -x ' . $file['filename'];
                    }
                  }
                }

                // Extract.
                if (!empty($extract)) {
                  $commands[] = array(
                    FALSE,
                    'CRIT',
                    'tar' . $extract . ' -f ' . $server->temp_archive,
                    $server->webroot,
                  );
                }
                else {
                  _deployer_log(t('No files/ found in archive on right level.'));
                  $deploy_context['failed'] = 1;
                }
              }

              // Webroot without files / settings.php.
              if ($deploy_context['restore_webroot']) {
                $commands = array_merge($commands, $std_commands);
              }

            }
          }

          // ===> Generic commands <===
          //
          // Print tag to webroot.
          $commands[] = array(
            FALSE,
            'CRIT',
            'echo \'' . $deploy_context['tag_str'] . '\' > .deployed_tag',
            $server->webroot,
          );
          // Set ownership of webroot.
          $commands[] = array(
            FALSE,
            'CRIT',
            'find . -mindepth 1 -maxdepth 1 -exec chown -R ' . $server->owner_user . ':' . $server->owner_group . ' "{}" \;',
            $server->webroot,
          );

          // Set mode on webroot.
          $commands[] = array(
            FALSE,
            'CRIT',
            'find . -mindepth 1 -maxdepth 1 -exec chmod -R 750 "{}" \;',
            $server->webroot,
          );

          // Set mode on drupal files folder.
          $commands[] = array(
            FALSE,
            'CRIT',
            'find sites -mindepth 2 -maxdepth 2 -type d -name files -exec chmod -R 770 "{}" \; -exec chmod 7770 "{}" \;',
            $server->webroot,
          );

          // Remove temp archive.
          $commands[] = array(
            TRUE,
            'WARN',
            'rm -f ' . $temp_files_archive,
            $server->webroot,
          );

          // Set drush commands.
          $field_revert_features = field_get_items('node', $deploynode, 'field_revert_features');
          $field_drush_clear_cache = field_get_items('node', $deploynode, 'field_drush_clear_cache');
          $field_drush_update = field_get_items('node', $deploynode, 'field_drush_update');
          $field_l10n_update = field_get_items('node', $deploynode, 'field_l10n_update');
          $field_search_api_reindex = field_get_items('node', $deploynode, 'field_search_api_reindex');

          foreach ($server->sites as $site) {
            // Set drush uri.
            if ($site == 'default') {
              $uri = '';
            }
            else {
              $uri = ' --uri=' . $site;
            }

            if ($field_drush_update[0]['value'] === 1) {
              $commands[] = array(
                FALSE,
                'WARN',
                $server->drush . $uri . ' updatedb -y',
                $server->webroot,
              );
            }
            if ($field_revert_features[0]['value'] === 1 && $field_drush_clear_cache[0]['value'] === 1) {
              $commands[] = array(
                FALSE,
                'WARN',
                $server->drush . $uri . ' cc all; ' . $server->drush . $uri . ' fra -y; ' . $server->drush . $uri . ' cc all; ' . $server->drush . $uri . ' fra -y; ' . $server->drush . $uri . ' cc all',
                $server->webroot,
              );
            }
            elseif ($field_revert_features[0]['value'] === 1) {
              $commands[] = array(
                FALSE,
                'WARN',
                $server->drush . $uri . ' fra -y',
                $server->webroot,
              );
            }
            elseif ($field_drush_clear_cache[0]['value'] === 1) {
              $commands[] = array(
                FALSE,
                'WARN',
                $server->drush . $uri . ' cc all',
                $server->webroot,
              );
            }
            if ($field_l10n_update[0]['value'] === 1) {
              $commands[] = array(
                FALSE,
                'WARN',
                $server->drush . $uri . ' l10n-update-refresh; ' . $server->drush . $uri . ' l10n-update',
                $server->webroot,
              );
            }
            if ($field_search_api_reindex[0]['value'] === 1) {
              $commands[] = array(
                FALSE,
                'WARN',
                $server->drush . $uri . ' drush sapi-c; ' . $server->drush . $uri . ' drush sapi-i',
                $server->webroot,
              );
            }
          }

          // Execute commands.
          _deployer_ssh_command($server, $commands, $deploy_context);

          $server->webroot_touched = 1;

        }
      }
    }
  }
}

/**
 * Load databases in case of restore.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_unpack_databases_ssh($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed.
  if (!$deploy_context['failed'] && $deploy_context['mode'] == 'restore') {
    if (isset($deploy_context['restores']['database_backup']) && !empty($deploy_context['restores']['database_backup'])) {

      _deployer_log("\n" . t('Restore routine for databases.'), NULL, DEPLOYER_NO_DEBUG);

      foreach ($servers as &$server) {

        // Rollback db's.
        foreach ($deploy_context['restores']['database_backup'] as $backup) {

          if (!isset($backup['site']) || !isset($server->temp_db_archive[$backup['site']]) || empty($server->temp_db_archive[$backup['site']])) {
            _deployer_log(t('Cannot rollback databases, no site or filename available! (CRITICAL)'));
            $deploy_context['failed'] = 1;
            continue;
          }

          // Get mysql connection string from settings.php.
          // We don't use drush here, since drush might fail bootstrapping
          // on an empty or invalid database, something which is a common
          // situation in case you want to do a restore.
          $cmd = 'php ' . $server->rpc . ' ' . $server->webroot . '/sites/' . $backup['site'] . '/settings.php';
          if ($server->username == 'root') {
            // If root, execute command as webroot owner.
            // Otherwise, we would execute unknown code as user, which mich
            // imply major security issues.
            $cmd = 'su ' . $server->owner_user . ' -c "' . $cmd . '"';
          }
          $commands = array(
            'connection' => array(
              FALSE,
              'CRIT',
              $cmd,
            ),
          );
          $result = _deployer_ssh_command($server, $commands, $deploy_context);

          if (!$deploy_context['failed'] && !empty($result['connection'])) {
            $mysql = 'mysql ' . $result['connection'];

            $commands = array();

            // Test archive integrity.
            $commands[] = array(
              FALSE,
              'CRIT',
              'gunzip -t ' . $server->temp_db_archive[$backup['site']],
              $server->webroot,
            );

            // Drop database tables.
            $commands[] = array(
              FALSE,
              'WARN',
              $mysql . ' -Nse "show tables" | while read table; do ' . $mysql . ' -e "drop table $table"; done',
              $server->webroot,
            );

            // Load backup archive.
            $commands[] = array(
              FALSE,
              'WARN',
              'zcat ' . $server->temp_db_archive[$backup['site']] . ' | ' . $mysql,
              $server->webroot,
            );

            // Execute commands.
            _deployer_ssh_command($server, $commands, $deploy_context);
          }
        }
      }
      _deployer_log(t('Restore routine for database done.'));
    }
  }
}

/**
 * Reset all user passwords and retrieve admin uli.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_reset_all_passwords_ssh($deploynode, &$deploy_context, &$servers) {

  if ($deploy_context['failed']) {
    _deployer_log('FAILED!!');
  }
  // Check if routine failed.
  if (!$deploy_context['failed'] && ($deploy_context['mode'] == 'restore' || $deploy_context['mode'] == 'deploy')) {
    if ($deploy_context['reset_passwords']) {
      _deployer_log("\n" . t('Resetting all drupal passwords.'), NULL, DEPLOYER_NO_DEBUG);

      foreach ($servers as &$server) {
        // Only run if version was retrieved.
        if ($server->current_drupal_version) {
          if (strpos($server->current_drupal_version, '7.') === 0) {
            $pass_hash = _password_crypt('sha512', _deployer_gen_uuid(), _password_generate_salt(DRUPAL_MIN_HASH_COUNT));
            $query = "echo 'update users set pass=\"$pass_hash\" where uid != 0;' | ";
          }
          elseif (strpos($server->current_drupal_version, '6.') === 0) {
            $pass_hash = md5(_deployer_gen_uuid());
            $query = "echo 'update users set pass=\"$pass_hash\" where uid != 0;' | ";
          }

          if (isset($query)) {
            $server->admin_reset_uri = array();
            foreach ($server->sites as $site) {
              // Set drush uri.
              if ($site == 'default') {
                $uri = '';
              }
              else {
                $uri = ' --uri=' . $site;
              }

              $commands = array();

              // Make all passwords unusable.
              $commands['reset'] = array(
                TRUE,
                'WARN',
                $query . $server->drush . $uri . ' sql-cli',
                $server->webroot,
              );

              // Remove all sessions.
              $commands['remove_sessions'] = array(
                TRUE,
                'WARN',
                "echo \"truncate table sessions;\" | " . $server->drush . $uri . ' sql-cli',
                $server->webroot,
              );

              // Get new admin reset link.
              $commands['uli'] = array(
                TRUE,
                'WARN',
                $server->drush . $uri . ' uli',
                $server->webroot,
              );

              // Action.
              $output = _deployer_ssh_command($server, $commands, $deploy_context);
              $server->admin_reset_uri[$site] = $output['uli'];
            }
          }
        }

        // Log if drupal version was not retrieved.
        if (!isset($query)) {
          _deployer_log("\n" . t('Cannot reset passwords on this webroot. Drupal version unknown or not retrieved. (WARNING)'), NULL, DEPLOYER_NO_DEBUG);
        }
      }
    }
  }
}

/**
 * Rollback webroot in case of failure.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_rollback_webroot_ssh($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed, otherwise no need for rollback.
  if ($deploy_context['failed']) {
    _deployer_log("\n" . t('Rollback routine for webroot.'), NULL, DEPLOYER_NO_DEBUG);

    foreach ($servers as &$server) {
      if ($deploy_context['webroot_backup']) {
        if ($server->webroot_touched) {
          if (isset($server->webroot_backupfile)) {

            $commands = array(
              // Test archive integrity.
              array(
                FALSE,
                'WARN',
                'tar -tf ' . $server->webroot_backupfile,
                $server->webroot,
              ),
              // Remove webroot including hidden files.
              array(
                FALSE,
                'WARN',
                'rm -rf *; for file in `find -type f`; do rm -f $file; done',
                $server->webroot,
              ),
              // Extract backup archive.
              array(
                FALSE,
                'WARN',
                'tar -xf ' . $server->webroot_backupfile . ' --overwrite --no-overwrite-dir',
                $server->webroot,
              ),
            );

            // Set failure temp off, to see if execution failed or not.
            $deploy_context['failed'] = 0;

            // Execute commands.
            _deployer_ssh_command($server, $commands, $deploy_context);

            if ($deploy_context['failed']) {
              _deployer_log(t('Rollback of webroot failed on server %h%! (CRITICAL)', array('%h%' => $server->id)), NULL, DEPLOYER_NO_DEBUG);
            }
            else {
              _deployer_log(t('Rollback of webroot ok on server %h%.', array('%h%' => $server->id)), NULL, DEPLOYER_NO_DEBUG);
              $deploy_context['failed'] = 1;
            }
          }
          else {
            _deployer_log(t('No backup file present, cannot no rollback on server %h%. (CRITICAL)', array('%h%' => $server->id)), NULL, DEPLOYER_NO_DEBUG);
          }
        }
        else {
          _deployer_log(t('Webroot not touched on server %h%, no need for rollback.', array('%h%' => $server->id)), NULL, DEPLOYER_NO_DEBUG);
        }
      }
      else {
        _deployer_log(t('You choose not to do any backup, so no rollback possible.'), NULL, DEPLOYER_NO_DEBUG);
      }
    }
  }
}

/**
 * Rollback databases in case of failure.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_rollback_database_ssh($deploynode, &$deploy_context, &$servers) {

  // Check if routine failed, otherwise no need for rollback.
  if ($deploy_context['failed']) {
    _deployer_log("\n" . t('Rollback routine for database.'), NULL, DEPLOYER_NO_DEBUG);

    if ($deploy_context['database_backup']) {
      foreach ($servers as &$server) {

        if ($server->database_touched || $server->webroot_touched) {
          // Rollback db's.
          foreach ($server->db_backups as $backup) {

            if (!isset($backup['site']) || !isset($backup['filename'])) {
              _deployer_log(t('Backup for this instance not present!! CANNOT ROLLBACK DATABASE! (CRITICAL)'));
              continue;
            }

            // Set drush uri to act on.
            if ($backup['site'] == 'default') {
              $uri = '';
            }
            else {
              $uri = ' --uri=' . $backup['site'];
            }

            $commands = array();

            // Test archive integrity.
            $commands[] = array(
              FALSE,
              'WARN',
              'gunzip -t ' . $backup['filename'],
              $server->webroot,
            );

            // Drop database tables.
            $commands[] = array(
              FALSE,
              'WARN',
              $server->drush . $uri . ' sql-drop -y',
              $server->webroot,
            );

            // Load backup archive.
            $commands[] = array(
              FALSE,
              'WARN',
              'zcat ' . $backup['filename'] . ' | ' . $backup['sql-connect'],
              $server->webroot,
            );

            // Set failure temp off, to see if execution failed or not.
            $deploy_context['failed'] = 0;

            // Execute commands.
            _deployer_ssh_command($server, $commands, $deploy_context);

            if ($deploy_context['failed']) {
              _deployer_log(t('Rollback of database for site %s% failed on server %h%! (CRITICAL)', array('%s%' => $backup['site'], '%h%' => $server->id)));
            }
            else {
              _deployer_log(t('Rollback of database for site %s% ok on server %h%.', array('%s%' => $backup['site'], '%h%' => $server->id)));
              $deploy_context['failed'] = 1;
            }

          }
        }
        else {
          _deployer_log(t('Database and webroot not touched on server %h%, so no need to rollback.', array('%h%' => $server->id)));
        }
      }
    }
    else {
      _deployer_log(t('You choose not to make a database backup, so rollback is impossible.'));
    }
    _deployer_log(t('Rollback routine for database done.'));
  }
}


/**
 * Finish the deploy, clean up archives etc.
 *
 * @param object $deploynode
 *   Deploy node object.
 * @param array $deploy_context
 *   Deploy context.
 * @param array $servers
 *   Array containing server objects.
 */
function deployer_batch_action_finish_ssh($deploynode, &$deploy_context, &$servers) {

  _deployer_log("\n" . t('Finishing deploy now.'), NULL, DEPLOYER_NO_DEBUG);

  $backup_meta_data = array();

  // Prepate backup location on deployer server.
  $create_path = drupal_realpath('private://deploy/backups');
  if (!file_prepare_directory($create_path)) {
    drupal_mkdir('private://deploy/backups', NULL, TRUE);
  }

  // Keep track of failed backups.
  $backups_ok = -1;

  // Save admin password reset uri if set.
  $admin_password_reset_uri = array();

  $server_counter = 0;
  foreach ($servers as &$server) {

    _deployer_log(t('Cleaning up on server %h%.', array('%h%' => $server->id)));

    $dels = array();
    $dels_folder = array();
    $gets = array();

    // Remove temp webroot archive.
    $dels[] = array('WARN', $server->temp_archive);

    // Remove database archive (restore) if any.
    if (!empty($server->temp_db_archive)) {
      foreach ($server->temp_db_archive as $db_archive) {
        $dels[] = array('WARN', $db_archive);
      }
    }

    // Remove RPC used for DB restores.
    if (isset($server->rpc)) {
      $dels[] = array('WARN', $server->rpc);
    }

    // Remove drush archive.
    if (isset($server->drush_archive)) {
      $dels[] = array('WARN', $server->drush_archive);
    }

    // Remove drush unpacked archive.
    if (isset($server->drush_folder)) {
      $dels_folder[] = array('WARN', $server->drush_folder);
    }

    // Download / remove webroot backup file.
    if (isset($server->webroot_backupfile)) {

      // Only first server in cluster.
      if ($server_counter == 0) {
        $server->webroot_local_backupfile = 'private://deploy/backups/' . uniqid() . '_webroot.tar.gz';

        $gets_now = array();
        $gets_now[] = array(
          'WARN',
          $server->webroot_backupfile,
          drupal_realpath($server->webroot_local_backupfile),
        );

        _deployer_sftp_get($server->sftp, $gets_now);

        // Backup meta data.
        if (is_file(drupal_realpath($server->webroot_local_backupfile))) {
          $backup_meta_data[$server_counter]['webroot_backup'] = $server->webroot_local_backupfile;
          $backup_meta_data[$server_counter]['tag_str'] = $server->tag_str;
          // Create file entity.
          _deployer_create_save_file_entity($server->webroot_local_backupfile);
          if ($backups_ok != 0) {
            $backups_ok = 1;
          }
        }
        else {
          $backups_ok = 0;
        }
      }

      $dels[] = array('WARN', $server->webroot_backupfile);
    }

    // Download / remove database backupfiles.
    if (isset($server->db_backups[0]) && !empty($server->db_backups[0]['filename'])) {

      $backup_counter = 0;
      foreach ($server->db_backups as $backup) {

        if (isset($backup['filename'])) {
          // Only first server in cluster.
          if ($server_counter == 0) {
            _deployer_log(t('Retrieving db backup for site %s% from server %h%.', array('%s%' => $backup['site'], '%h%' => $server->id)));
            $backup['localfile'] = 'private://deploy/backups/' . uniqid() . '_db.sql.gz';

            $gets_now = array();
            $gets_now[] = array(
              'WARN',
              $backup['filename'],
              drupal_realpath($backup['localfile']),
            );

            _deployer_sftp_get($server->sftp, $gets_now);

            // Backup meta data.
            if (is_file(drupal_realpath($backup['localfile']))) {
              $backup_meta_data[$server_counter]['database_backup'][$backup_counter]['site'] = $backup['site'];
              $backup_meta_data[$server_counter]['database_backup'][$backup_counter]['localfile'] = $backup['localfile'];

              // Create file entity.
              _deployer_create_save_file_entity($backup['localfile']);
              if ($backups_ok != 0) {
                $backups_ok = 1;
              }
            }
            else {
              $backups_ok = 0;
            }

            _deployer_log(t('Backup status: @s', array('@s' => $backups_ok)));
            $server->db_backups[$backup_counter] = $backup;
          }

          $dels[] = array(
            'WARN',
            $backup['filename'],
          );
        }

        $backup_counter++;
      }
    }

    // Save admin password reset uri if set.
    if (isset($server->admin_reset_uri)) {
      $admin_password_reset_uri[] = array(
        'hostname' => $server->hostname,
        'webroot' => $server->webroot,
        'uri' => $server->admin_reset_uri,
      );
    }

    _deployer_sftp_get($server->sftp, $gets);
    _deployer_sftp_delete_file($server->sftp, $dels);
    _deployer_sftp_delete_folder($server->sftp, $dels_folder);

    $server_counter++;
  }

  // Save backup metadata.
  if (!empty($backup_meta_data)) {
    _deployer_log(t('Saving backup metadata.'));
    file_put_contents(drupal_realpath('private://deploy/backups/' . $deploynode->nid . '.meta'), drupal_json_encode($backup_meta_data));
  }

  // Set deploystatus on node to 1 (finished) and save backup status.
  $temp_node = node_load($deploynode->nid, $vid = NULL, $reset = TRUE);

  if ($deploy_context['failed']) {
    $temp_node->field_deploystatus[LANGUAGE_NONE][0]['value'] = 2;
  }
  else {
    $temp_node->field_deploystatus[LANGUAGE_NONE][0]['value'] = 1;
  }

  if ($backups_ok <= 0) {
    $backups_ok = 0;
  }
  $temp_node->field_backups_ok[LANGUAGE_NONE][0]['value'] = $backups_ok;
  _deployer_log(t('Backup status: @s', array('@s' => $backups_ok)));

  // Save admin password reset uri if set.
  if (!empty($admin_password_reset_uri)) {
    $admin_password_reset_uri_var = variable_get('deployer_admin_reset_uri', array());
    $admin_password_reset_uri_var[$deploy_context['env_id']] = array(
      'uri' => $admin_password_reset_uri,
      'time' => time(),
    );
    variable_set('deployer_admin_reset_uri', $admin_password_reset_uri_var);
  }

  _deployer_log(t('Triggering health-status update on environment...'), NULL, DEPLOYER_NO_DEBUG);
  _deployer_server_checkup_ssh_act($deploy_context['environment']->nid);
  _deployer_log(t('Finished.'), NULL, DEPLOYER_NO_DEBUG);
  node_save($temp_node);
}
