<?php
/**
 * @file
 * Show deploy information after it has happened.
 */

/**
 * Implements hook_node_view().
 */
function deployer_node_view($node, $view_mode, $langcode) {
  if ($node->type == 'deploy') {

    if ($node->status == 1) {
      // Check if deploy seems still running, it actually IS
      // still running by checking on the lockfile.
      // If the lockfile is still there, it's ok.
      // If not, the process ended and the lockfile was removed.
      // That means that de deploy status is out of sync,
      // so we set it to 'failed'.
      if (deployer_deploy_is_running($node)) {
        $deploy_crashed = FALSE;
        $lock = new DeployerLockNode($node->nid);
        if (!$lock->hasLock()) {
          // Not locked, but still running?  hmmmmm....
          // Retrieve fresh node object, so we are sure that the lockfile is
          // not just deleted in the meantime, on a natural way.
          $fresh_node = node_load($node->nid, NULL, $reset = TRUE);
          // Check the status again.
          if (deployer_deploy_is_running($fresh_node)) {
            $deploy_crashed = TRUE;
          }
        }
        else {
          // Lock still exists, but is its process still running?
          if (!$lock->lockProcAlive()) {
            $deploy_crashed = TRUE;
          }
        }

        // Deploy really is not running anymore.
        // So status is out of sync.
        if ($deploy_crashed) {
          // Remove lock file.
          $lock->unlock();
          // Set deploy status to failed.
          $fresh_node = node_load($node->nid);
          $fresh_node->field_deploystatus[LANGUAGE_NONE][0]['value'] = 2;
          // Append crash-warning to lofile.
          _deployer_log(t('This deploy crashed somehow. I unlocked it. Check your environment!'), $fresh_node, $debug = FALSE, $replace = FALSE);
          // Save node and refresh page.
          node_save($fresh_node);
          drupal_goto('node/' . $fresh_node->nid);
        }
      }

      // Replace deploy-status field with human readable text.
      // See _deployer_interpret_deploystatus() for status info / meaning.
      $status = _deployer_interpret_deploystatus($node->content['field_deploystatus'][0]['#markup']);
      $node->content['field_deploystatus'][0]['#markup'] = '<span class="' . drupal_strtolower(str_replace(' ', '', $status)) . '">' . $status . '</span>';

      $field_deploystatus = field_get_items('node', $node, 'field_deploystatus');
      $field_deploystatus = $field_deploystatus[0]['value'];
      if ($view_mode == 'full') {
        if ($field_deploystatus < 1) {
          // Deploy is running now.
          // Add ajax to tail log.
          drupal_add_js(drupal_get_path('module', 'deployer') . '/js/getlog.js');
          drupal_add_js(array('deployer' => array('deploy_id' => $node->nid)), 'setting');
        }
        else {
          // Only add js to hide / show debug lines.
          drupal_add_js(drupal_get_path('module', 'deployer') . '/js/hideDebug.js');
        }
      }
    }

    // Remove information out of scope.
    $restore_id = field_get_items('node', $node, 'field_restore_id');
    if (isset($restore_id[0]['value'])) {
      unset($node->content['#groups']['group_deploy_options']);
      unset($node->content['#fieldgroups']['group_deploy_options']);
    }
    else {
      unset($node->content['#groups']['group_restore_options']);
      unset($node->content['#fieldgroups']['group_restore_options']);
    }

    // Show admin password reset links if any.
    $env_id = field_get_items('node', $node, 'field_environment_id');
    if (isset($env_id[0]['target_id'])) {
      $env_id = $env_id[0]['target_id'];
      $admin_password_reset_uri = variable_get('deployer_admin_reset_uri', '');
      if (!empty($admin_password_reset_uri) && isset($admin_password_reset_uri[$env_id])) {
        foreach ($admin_password_reset_uri[$env_id]['uri'] as $uri_var) {
          $uri = $uri_var;
          foreach ($uri['uri'] as $sitename => $site) {
            $path = parse_url($site);
            if (!empty($path['query'])) {
              $path = $path['path'] . '?' . $path['query'];
            }
            else {
              $path = $path['path'];
            }
            drupal_set_message(t('Reset admin password for site !site (!host:!webroot) at !uri', array(
              '!site' => $sitename,
              '!host' => $uri['hostname'],
              '!webroot' => $uri['webroot'],
              '!uri' => $path,
            )));
          }
        }
      }
    }
  }
}

/**
 * Callback function to get deploy log content.
 *
 * @param int $deploy_id
 *   Node if of deploy node.
 */
function deployer_get_logfile_json($deploy_id) {
  $node = node_load($deploy_id);
  if ($node && $node->type == 'deploy') {
    $deploystatus = field_get_items('node', $node, 'field_deploystatus');
    $deploystatus = $deploystatus[0]['value'];

    if ($deploystatus < 1) {
      // Still running or idle.

      // Check for lock file.
      // It's will be removed on php shutdown, so we can see
      // if a deploy crashed or not.
      $lock = new DeployerLockNode($deploy_id);
      if (!$lock->hasLock()) {
        // Not running anymore!
        // Trigger refresh.
        echo 'FINISHED';
        drupal_exit();
      }

      $retry = 0;
      while ($retry <= 3) {

        // Get fresh node from DB.
        if ($retry > 0) {
          $node = node_load($deploy_id, NULL, $reset = TRUE);
        }

        $logfile = field_get_items('node', $node, 'field_logfile');
        if (!empty($logfile)) {
          $logfile = $logfile[0]['uri'];
          $logfile = drupal_realpath($logfile);
          if (file_exists($logfile)) {
            echo file_get_contents($logfile);
          }
          else {
            echo "Logfile not present.";
          }
          drupal_exit();
          break;
        }
        $retry++;
        sleep(1);
      }
      echo "FINISHED";
      drupal_exit();
    }
    else {
      echo "FINISHED";
      drupal_exit();
    }
  }
  echo 'FALSE';
  drupal_exit();
}

/**
 * Display health for server.
 *
 * @param object $field
 *   Field object.
 *
 * @return string
 *   Health information html.
 */
function deployer_show_server_health_full($field) {

  $cid = $field['entity']->item_id;

  $query = db_select('deployer_health_info', 's');
  $query->condition('s.cid', $cid);
  $query->fields('s', array(
    'cid',
    'alive',
    'is_drupal',
    'version',
    'updates',
    'deployed_version',
  ));
  $results = $query->execute();

  $variables['title'] = t('Health information');
  $variables['type'] = 'ul';
  $variables['attributes'] = array();
  $variables['items'] = array();

  $rowcount = $results->rowCount();
  if (empty($rowcount)) {
    $variables['title'] = t('Health information not available');
  }
  else {
    foreach ($results as $row) {

      // Server alive.
      if ($row->alive) {
        $alive = t('running');
      }
      else {
        $alive = t('down');
      }
      $variables['items'][] = t('Server status:') . ' ' . $alive;

      // Is drupal webroot.
      if ($row->is_drupal) {
        $drupal = t('yes');
      }
      else {
        $drupal = t('no');
      }
      $variables['items'][] = t('Drupal webroot:') . ' ' . $drupal;

      // Running version tag.
      $deployed_version_parts = unserialize(base64_decode($row->deployed_version));
      if (!empty($deployed_version_parts)) {
        $variables['items'][] = t('Running version / tag:') . ' ' . l($deployed_version_parts['tag'], $deployed_version_parts['url']);
      }

      // Drupal version.
      if (empty($row->version)) {
        $version = t('unknown');
      }
      else {
        $version = $row->version;
      }
      $variables['items'][] = t('Drupal version:') . ' ' . $version;

      // Updates.
      $updates = unserialize($row->updates);
      if (!empty($updates)) {
        $update_array['data'] = t('Updates');
        foreach ($updates as $update) {
          $mod_array = array();
          $mod_array['data'] = l($update['module'], 'https://drupal.org/project/' . $update['module']);
          $mod_array['children'][] = t('Current version:') . ' ' . $update['version'];
          $mod_array['children'][] = t('New version:') . ' ' . $update['new_version'];
          $mod_array['children'][] = t('Type:') . ' ' . $update['type'];

          $update_array['children'][] = $mod_array;
        }
        $variables['items'][] = $update_array;
      }
      else {
        $variables['items'][] = t('No updates');
      }

      break;
    }
  }

  return theme_item_list($variables);
}

/**
 * Show log for deploy in displaysuite.
 *
 * @param object $field
 *   field object (display suite).
 *
 * @return string
 *   log html.
 */
function deployer_show_deploy_log($field) {
  $node = $field['entity'];
  $logfile = field_get_items('node', $node, 'field_logfile');
  if (!empty($logfile)) {
    $logfile = $logfile[0]['uri'];
    if (!empty($logfile) && is_file(drupal_realpath($logfile))) {
      $log = file_get_contents(drupal_realpath($logfile));
    }
  }

  if (!isset($log) || !$log) {
    $log = ' Logfile not present on disk! ';
  }
  return '<pre id="deploy-log">' . $log . '</pre>';
}

/**
 * Show info for deploy in displaysuite field.
 *
 * @param object $field
 *   field object (display suite).
 *
 * @return string
 *   log html.
 */
function deployer_show_deploy_info_field($field) {
  $node = $field['entity'];
  return deployer_show_deploy_info($node->nid);
}

/**
 * Show info for restore in displaysuite field.
 *
 * @param object $field
 *   field object (display suite).
 *
 * @return string
 *   log html.
 */
function deployer_show_restore_info_field($field) {
  $node = $field['entity'];
  $restore_id = field_get_items('node', $node, 'field_restore_id');
  if ($restore_id) {
    $restore_id = $restore_id[0]['value'];
    return deployer_show_deploy_info($restore_id, 'Restore information');
  }
  else {
    return FALSE;
  }
}

/**
 * Return backup link for environment.
 *
 * @param object $field
 *   Field object, containing node.
 *
 * @return string
 *   HTML.
 */
function deployer_env_backup_link($field) {
  $node = $field['entity'];
  if ($node->type == 'environment') {
    return l(t('Backup'), 'node/add/deploy', array('query' => array('env' => $node->nid, 'backup' => 1)));
  }

  return FALSE;
}

/**
 * Return deploy link for environment.
 *
 * @param object $field
 *   Field object, containing node.
 *
 * @return string
 *   HTML.
 */
function deployer_env_deploy_link($field) {
  $node = $field['entity'];
  if ($node->type == 'environment') {
    return l(t('Deploy'), 'node/add/deploy', array('query' => array('env' => $node->nid)));
  }
  return FALSE;
}


/**
 * Return connected projects output for client.
 *
 * @param object $field
 *   Field object.
 *
 * @return string|bool
 *   HTML output or FALSE.
 */
function deployer_client_connected_projects($field) {
  return views_embed_view('projects', 'block', $field['entity']->nid);
}

/**
 * See if deploy is running.
 *
 * @param object $node
 *   Node object containing field field_deploystatus.
 *
 * @return bool
 *   TRUE if deploy seems running, FALSE if not (idle of finished).
 */
function deployer_deploy_is_running($node) {
  $deploy_status = field_get_items('node', $node, 'field_deploystatus');
  if (!empty($deploy_status)) {
    $deploy_status = $deploy_status[0]['value'];
    if ($deploy_status == 0.5) {
      return TRUE;
    }
  }
  return FALSE;
}
